from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def tensor_proc(labels_dict,input_dir):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    output_dir = cwd
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')
    if not session_label:
        search_session_label="*"
    else:
        search_session_label=session_label

    input_dir=substitute_labels(input_dir,labels_dict)
    dwi = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(search_session_label),'dwi','*_desc-preproc_dwi.nii.gz'))
    bval = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(search_session_label),'dwi','*_desc-preproc_dwi.bval'))
    bvec = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(search_session_label),'dwi','*_desc-preproc_dwi.bvec'))
    mask = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(search_session_label),'dwi','*_desc-brain_mask.nii.gz'))
    grad = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(search_session_label),'dwi','*_desc-preproc_dwi.b'))

    tensor_dir = os.path.join(cwd,'tensors')
    if not os.path.isdir(tensor_dir):
        os.makedirs(tensor_dir)

    if not session_label:
        session_label=get_bidstag("ses",os.path.basename(dwi))
        if not session_label:
            tensor_mrtrix = os.path.join(tensor_dir,'sub-{}_tensor.mif'.format(participant_label))
            kurtosis_mrtrix = os.path.join(tensor_dir,'sub-{}_kurtosis.mif'.format(participant_label))
        else:
            tensor_mrtrix = os.path.join(tensor_dir,'sub-{}_ses-{}_tensor.mif'.format(participant_label,session_label))
            kurtosis_mrtrix = os.path.join(tensor_dir,'sub-{}_ses-{}_kurtosis.mif'.format(participant_label,session_label))           
    else:
        tensor_mrtrix = os.path.join(tensor_dir,'sub-{}_ses-{}_tensor.mif'.format(participant_label,session_label))
        kurtosis_mrtrix = os.path.join(tensor_dir,'sub-{}_ses-{}_kurtosis.mif'.format(participant_label,session_label))

    command_base, container = getContainer(labels_dict,nodename="tensor", SPECIFIC="MRTRIX_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the dwi2tensor version:")
    command = f"{command_base} dwi2tensor --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    if container:
        IFLOGGER.info("\nChecking the container version:")
        command = f"{command_base} --version"
        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

    # generate tensors using mrtrix, use grad from qsiprep
    params="-grad "+grad+\
        " -mask "+mask+\
        " -dkt "+kurtosis_mrtrix+\
        " "+dwi+\
        " "+tensor_mrtrix

    command=f"{command_base} dwi2tensor"\
            " "+params

    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    
    # generate tensor metrics
    tensor_metrics_dir = os.path.join(cwd,'tensor_metrics')
    if not os.path.isdir(tensor_metrics_dir):
        os.makedirs(tensor_metrics_dir)

    if not session_label:
        session_label=get_bidstag("ses",os.path.basename(dwi))
        if not session_label:
            fa_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_space-T1w_desc-preproc_desc-fa.mif'.format(participant_label))
            adc_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_space-T1w_desc-preproc_desc-adc.mif'.format(participant_label))
            ad_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_space-T1w_desc-preproc_desc-ad.mif'.format(participant_label))
            rd_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_space-T1w_desc-preproc_desc-rd.mif'.format(participant_label))   
        else:
            fa_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-fa.mif'.format(participant_label,session_label))
            adc_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-adc.mif'.format(participant_label,session_label))
            ad_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-ad.mif'.format(participant_label,session_label))
            rd_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-rd.mif'.format(participant_label,session_label))      
    else:
        fa_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-fa.mif'.format(participant_label,session_label))
        adc_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-adc.mif'.format(participant_label,session_label))
        ad_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-ad.mif'.format(participant_label,session_label))
        rd_mrtrix = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-rd.mif'.format(participant_label,session_label))


    # generate tensors using mrtrix
    params=" -mask "+mask+\
        " -fa "+fa_mrtrix+\
        " -adc "+adc_mrtrix+\
        " -ad "+ad_mrtrix+\
        " -rd "+rd_mrtrix+\
        " "+tensor_mrtrix

    command=f"{command_base} tensor2metric"\
            " "+params

    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    # convert to nifti
    if not session_label:
        session_label=get_bidstag("ses",os.path.basename(dwi))
        if not session_label:
            fa_fsl = os.path.join(tensor_metrics_dir,'sub-{}_space-T1w_desc-preproc_desc-fa.nii.gz'.format(participant_label))
            adc_fsl = os.path.join(tensor_metrics_dir,'sub-{}_space-T1w_desc-preproc_desc-adc.nii.gz'.format(participant_label))
            ad_fsl = os.path.join(tensor_metrics_dir,'sub-{}_space-T1w_desc-preproc_desc-ad.nii.gz'.format(participant_label))
            rd_fsl = os.path.join(tensor_metrics_dir,'sub-{}_space-T1w_desc-preproc_desc-rd.nii.gz'.format(participant_label))  
        else:
            fa_fsl = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-fa.nii.gz'.format(participant_label,session_label))
            adc_fsl = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-adc.nii.gz'.format(participant_label,session_label))
            ad_fsl = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-ad.nii.gz'.format(participant_label,session_label))
            rd_fsl = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-rd.nii.gz'.format(participant_label,session_label))     
    else:
        fa_fsl = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-fa.nii.gz'.format(participant_label,session_label))
        adc_fsl = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-adc.nii.gz'.format(participant_label,session_label))
        ad_fsl = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-ad.nii.gz'.format(participant_label,session_label))
        rd_fsl = os.path.join(tensor_metrics_dir,'sub-{}_ses-{}_space-T1w_desc-preproc_desc-rd.nii.gz'.format(participant_label,session_label)) 


    # generate tensors using mrtrix
    params=fa_mrtrix+" "+ fa_fsl
    command=f"{command_base} mrconvert"\
            " "+params
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    params=adc_mrtrix+" "+ adc_fsl
    command=f"{command_base} mrconvert"\
            " "+params
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    params=ad_mrtrix+" "+ ad_fsl
    command=f"{command_base} mrconvert"\
            " "+params
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    params=rd_mrtrix+" "+ rd_fsl
    command=f"{command_base} mrconvert"\
            " "+params
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    out_files=[]
    out_files.insert(0,fa_fsl)
    out_files.insert(1,adc_fsl)
    out_files.insert(2,ad_fsl)
    out_files.insert(3,rd_fsl)


    return {
        "fa":fa_fsl,
        "adc":adc_fsl,
        "ad":ad_fsl,
        "rd":rd_fsl,
        "output_dir":output_dir,
        "out_files":out_files
    }



class tensorInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    input_dir = traits.String("",desc="QSIPREP Output Directory", usedefault=True)

class tensorOutputSpec(TraitedSpec):
    fa = File(desc='FA')
    adc = File(desc='ADC')
    ad = File(desc='AD')
    rd = File(desc='RD')
    output_dir = traits.String(desc="NODDI output directory")
    out_files = traits.List(desc='list of files')
    
class tensor_pan(BaseInterface):
    input_spec = tensorInputSpec
    output_spec = tensorOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = tensor_proc(
            self.inputs.labels_dict,
            self.inputs.input_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="tensor_node",input_dir="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(tensor_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if input_dir is None:
        input_dir = ""
        
    pan_node.inputs.input_dir =  input_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def dummy_proc(labels_dict,bids_dir=""):

    DEBUG=True

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)


    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["SINGULARITYENV_TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME

    command_base, container = getContainer(labels_dict,nodename="qsiprep", SPECIFIC="QSIPREP_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the qsiprep version:")
    command = f"{command_base} --version"
    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    params="--participant_label <PARTICIPANT_LABEL>" \
        " --separate-all-dwis"\
        " --hmc-model eddy"\
        " --eddy-config <EDDY_CONFIG>" \
        " --unringing-method mrdegibbs" \
        " --mem_mb <BIDSAPP_MEMORY>" \
        " --nthreads <BIDSAPP_THREADS>"\
        " --fs-license-file <FSLICENSE>"\
        " --skip-bids-validation"\
        " -w <CWD>/qsiprep_work"\
        " --write-graph"\
        " --output-resolution <OUTPUT_RES>"

    command=f"{command_base}"\
            " "+bids_dir +\
            " <CWD>"\
            " participant"\
            " "+params


    evaluated_command=substitute_labels(command, labels_dict)
    IFLOGGER.info(evaluated_command)
    evaluated_command_args = shlex.split(evaluated_command)
    IFLOGGER.info(evaluated_command_args)

    if not DEBUG:
        results = subprocess.run(evaluated_command_args)
        IFLOGGER.info(results.stdout)

    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    dwi_preprocess = getGlob(os.path.join(cwd,'qsiprep','sub-{}'.format(participant_label),'ses-*','dwi','*preproc_dwi.nii.gz'))
    mat_t12mni = getGlob(os.path.join(cwd,'qsiprep','sub-{}'.format(participant_label),'anat','*from-T1w*mode-image_xfm.h5'))
    mat_mni2t1 = getGlob(os.path.join(cwd,'qsiprep','sub-{}'.format(participant_label),'anat','*from-MNI*mode-image_xfm.h5'))
    t1ref =  getGlob(os.path.join(cwd,'qsiprep','sub-{}'.format(participant_label),'anat','*{}_desc-preproc_T1w.nii.gz'.format(participant_label)))
    mniref =  getGlob(os.path.join(cwd,'qsiprep','sub-{}'.format(participant_label),'anat','*{}*MNI*desc-preproc_T1w.nii.gz'.format(participant_label)))
    output_dir = cwd

    
    out_files=[]
    out_files.insert(0,dwi_preprocess)
    out_files.insert(1,mat_t12mni)
    out_files.insert(2,mat_mni2t1)
    out_files.insert(3,mniref)
    out_files.insert(4,t1ref)


    return {
        "dwi_preprocess":dwi_preprocess,
        "mat_t12mni":mat_t12mni,
        "mat_mni2t1":mat_mni2t1,
        "mniref":mniref,
        "t1ref":t1ref,
        "output_dir":output_dir,
        "out_files":out_files
    }



class dummyInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class dummyOutputSpec(TraitedSpec):
    dwi_preprocess = File(desc='Preprocessed DWI')
    mat_t12mni = File(desc='T1 to MNI transform')
    mat_mni2t1 = File(desc='MNI to T1 transform')
    mniref = File(desc='MNI reference')
    t1ref = File(desc='T1 reference')
    output_dir = traits.String(desc="QSIPREP output directory")
    out_files = traits.List(desc='list of files')
    
class dummy_pan(BaseInterface):
    input_spec = dummyInputSpec
    output_spec = dummyOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = dummy_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='dummy_node',bids_dir="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(dummy_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





from nipype import Workflow, MapNode, Node

import panpipelines.nodes.antstransform as antstransform
import panpipelines.nodes.parse_textdata as parse_textdata
from panpipelines.utils.util_functions import *
from panpipelines.utils.transformer import *
import glob

def create(name, wf_base_dir,labels_dict,createGraph=True,execution={},LOGGER=None):
    # Create workflow
    pan_workflow = Workflow(name=name, base_dir=wf_base_dir)

    if LOGGER:
        LOGGER.info(f"Created Workflow {name} with base directory {wf_base_dir}")

    if len(execution.keys()) > 0:
        pan_workflow.config = process_dict(pan_workflow.config,execution)


    parsetextdata_node = parse_textdata.create(labels_dict,name="subject_text",LOGGER=LOGGER)
    parsetextdata_map_node = MapNode(parsetextdata_node.interface,name="subject_text_map",iterfield=['textdata'])

    measures_list=[]
    measures_template = getParams(labels_dict,"MEASURES_TEMPLATE")
    if isinstance(measures_template,list):
        for meas_template in measures_template:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict)
            measures_list.extend(glob.glob(evaluated_meas_template))
    else:
        measures_list.extend(glob.glob(measures_template))
        
    measures_list.sort()

    parsetextdata_map_node.inputs.textdata = measures_list
    pan_workflow.add_nodes([parsetextdata_map_node])


    if createGraph:
         pan_workflow.write_graph(graph2use='flat')


    return pan_workflow



#!/usr/bin/env python
from panpipelines.utils.util_functions import *
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from functools import partial
import json
import pandas as pd
import os
import logging
import sys
import multiprocessing as mp
from functools import partial
from panpipelines.version import __version__
from panpipelines.single_subject import runSingleSubject
from panpipelines.group_subjects import runGroupSubjects

LOGGER = logger_setup("panpipelines", logging.DEBUG)
logger_addstdout(LOGGER, logging.INFO)

PROCESSING_OPTIONS=["slurm", "local"]


def parse_params():
    parser = ArgumentParser(description="Pan pipelines")
    PathExists = partial(path_exists, parser=parser)
    parser.add_argument("config_file", type=PathExists, help="Pipeline Configuration File")
    parser.add_argument("--pipeline_outdir", help="directory to place pipeline output")
    parser.add_argument("--participants_file", type=PathExists, help="list of participants")
    parser.add_argument("--sessions_file", type=PathExists, help="Comprehensive list of participants and sessions")
    parser.add_argument("--pipelines", nargs="+")
    parser.add_argument("--pipeline_match", nargs="+")
    parser.add_argument("--projects", nargs="+")
    parser.add_argument("--participant_label", nargs="*", type=drop_sub, help="filter by subject label (the sub- prefix can be removed).")
    parser.add_argument("--participant_exclusions", nargs="*", type=drop_sub, help="filter by subject label (the sub- prefix can be removed).")
    parser.add_argument("--session_label", nargs="*", type=drop_ses, help="filter by session label (the ses- prefix can be removed).")
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))
    return parser


def main():
    datelabel = datetime.datetime.now().strftime("%Y%m%d_%H%M%S_%f")

    job_ids={}

    parser=parse_params()
    args, unknown_args = parser.parse_known_args()

    panpipe_labels={}

    #  TO DO:
    #  In general need to think through how config variables are pulled in and overwritten at the 
    #  general and pipeline level. Current approach leaves a lot to be desired.
    #  We are lucky that pipeline_dir is really only referenced within the pipelines themselves and so we
    #  can get away with this here.
    #
    #  Envisage an approach whereby variables are referenced by a pipeline from both general and pipeline specific
    #  Using keyword in dictionary.
    #
    pipeline_outdir = args.pipeline_outdir
    if pipeline_outdir:
        pipeline_outdir = os.path.abspath(pipeline_outdir)
        panpipe_labels = updateParams(panpipe_labels,"PIPELINE_DIR",pipeline_outdir)

    panpipeconfig_file=str(args.config_file)
    panpipeconfig_json=None
    if os.path.exists(panpipeconfig_file):
        with open(panpipeconfig_file, 'r') as infile:
            panpipeconfig_json = json.load(infile)
    # obtain labels but do not process yet
    panpipe_labels = process_labels(panpipeconfig_json,panpipeconfig_file,panpipe_labels,insert=True, postpone=True)

    if __file__:
        pkgdir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
        panpipe_labels = insertParams(panpipe_labels,'PKG_DIR',pkgdir)
    panpipe_labels = insertParams(panpipe_labels,'PWD',str(os.getcwd()))

   # only need date label if pipeline outdir not provided. This helps with nipype caching for repeat runs
    if not pipeline_outdir:
        pipeline_outdir_suffix = "_" + datelabel
        pipeline_outdir_config = getParams(panpipe_labels,"PIPELINE_DIR")
        if pipeline_outdir_config:
            panpipe_labels = updateParams(panpipe_labels,"PIPELINE_DIR",pipeline_outdir_config + pipeline_outdir_suffix)
        else:
            pipeline_outdir_config = os.path.join(os.getcwd(),"pan_output" +  pipeline_outdir_suffix) 
            panpipe_labels = updateParams(panpipe_labels,"PIPELINE_DIR",pipeline_outdir_config)


    # Now we can resolve all the references based on precedence
    panpipe_labels = update_labels(panpipe_labels)

    pipeline_outdir=os.path.abspath(getParams(panpipe_labels,"PIPELINE_DIR"))
    if not os.path.exists(pipeline_outdir):
        os.makedirs(pipeline_outdir,exist_ok=True)

    # reproducible and slightly more robust name to faciliate nipype caching
    panlabel=os.path.basename(pipeline_outdir)

    LOGFILE=os.path.join(pipeline_outdir,f"{panlabel}_pan_processing.log")
    logger_addfile(LOGGER, LOGFILE, logging.DEBUG)

    LOGGER.info(f"Running {__file__} v{__version__}")
    LOGGER.info("Running Pan Processing")
    LOGGER.info(f"start logging to {LOGFILE}")

    label_key="CONFIG_FILE"
    label_value=panpipeconfig_file
    panpipe_labels = updateParams(panpipe_labels, label_key,label_value)

    credentials = os.path.abspath(getParams(panpipe_labels,"CREDENTIALS"))
    if credentials is not None and os.path.exists(credentials):
        with open(credentials, 'r') as infile:
            cred_dict = json.load(infile)
            cred_user = getParams(cred_dict,"user")
            cred_password = getParams(cred_dict,"password")
    else:
        cred_user = "dummy_user"
        cred_password = "dummy_password"

    participants_file = args.participants_file
    if args.participants_file is not None:
        participants_file=str(args.participants_file)
        label_key="PARTICIPANTS_FILE"
        label_value=participants_file
        panpipe_labels = updateParams(panpipe_labels, label_key,label_value)
    else:
        participants_file = getParams(panpipe_labels,"PARTICIPANTS_FILE")

    projects=args.projects
    if not projects:
        projects=["001_HML","002_HML","003_HML","004_HML"]
        LOGGER.info(f"--projects parameter not set. Using defaults {projects} to retrieve subjects and sessions from XNAT.")
    else:
        LOGGER.info(f"Retrieving subjects and sessions from XNAT project {projects}")

    TEMPLATEFLOW_HOME=getParams(panpipe_labels,"TEMPLATEFLOW_HOME")
    if TEMPLATEFLOW_HOME:
        LOGGER.info(f"Downloading an initial set of TemplateFlow templates for spaces MNI152NLin2009cAsym and MNI152NLin6Asym to {TEMPLATEFLOW_HOME}.")
        initTemplateFlow(TEMPLATEFLOW_HOME) 

    # if participants file doesn't exist then lets download it
    if not os.path.exists(participants_file):
        LOGGER.info(f"Participants file not found at {participants_file} - retrieving from XNAT. Please wait.")
        xnat_host = getParams(panpipe_labels,"XNAT_HOST")
        targetdir = os.path.dirname(participants_file)
        if not os.path.exists(targetdir):
            os.makedirs(targetdir)
        getBidsTSV(xnat_host,cred_user,cred_password,projects,"BIDS-AACAZ",targetdir,demographics=False)

    sessions_file = args.sessions_file
    if args.sessions_file is not None:
        sessions_file=str(args.sessions_file)
        label_key="SESSIONS_FILE"
        label_value=sessions_file
        panpipe_labels = updateParams(panpipe_labels, label_key,label_value)
    else:
        sessions_file = getParams(panpipe_labels,"SESSIONS_FILE")

    pipelines=args.pipelines
    if args.pipelines is not None:
        label_key="PIPELINES"
        label_value=pipelines
        panpipe_labels = updateParams(panpipe_labels, label_key,label_value)
    else:
        pipelines = getParams(panpipe_labels,"PIPELINES")

    pipeline_match=args.pipeline_match

    if not pipelines:
        LOGGER.info("No pipelines specified at command line.")
        pipelines = [p for p in panpipeconfig_json.keys() if p != "all_pipelines"]
        if pipeline_match:
            LOGGER.info(f"All pipelines matching {str(pipeline_match)} in configuration file will be run.")
            pipeline_select=[]
            for pipe_match in pipeline_match:
                pipeline_select.extend([p for p in pipelines if pipe_match in p])
            pipelines = pipeline_select
        else:
            LOGGER.info("All pipelines in configuration file will be run.")

    # Remove duplicates in pipeline list
    pipelines = list(set(pipelines))

    LOGGER.info(f"About to arrange pipelines by dependency. Pipeline list is {pipelines}")    
    pipelines = arrangePipelines(panpipeconfig_json,pipelines=pipelines)
    LOGGER.info(f"Pipelines arranged by dependency. Pipeline list is {pipelines}")    

    participant_label = args.participant_label
    if args.participant_label is not None:
        label_key="PARTICIPANTS"
        label_value=participant_label
        panpipe_labels = updateParams(panpipe_labels, label_key,label_value)
    else:
        participant_label = getParams(panpipe_labels,"PARTICIPANTS")

    participant_exclusions = args.participant_exclusions
    if not participant_exclusions:
        participant_exclusions = []

    session_label = args.session_label
    if args.session_label is not None:
        label_key="SESSION_LABEL"
        label_value=session_label
        panpipe_labels = updateParams(panpipe_labels, label_key,label_value)
    else:
        session_label = getParams(panpipe_labels,"SESSION_LABEL")


    LOGGER.info(f"Pipelines to be processed : {pipelines}")

    projectmap = get_projectmap(participant_label, participants_file,session_labels=session_label,sessions_file=sessions_file,subject_exclusions=participant_exclusions)
    participant_list = projectmap[0]
    project_list  = projectmap[1]
    session_list = projectmap[2]

    # take snapshot of the runtime labels for all pipelines
    runtime_labels = panpipe_labels.copy()

    for pipeline in pipelines:
        LOGGER.info(f"Processing pipeline : {pipeline}")
        panpipe_labels = updateParams(panpipe_labels, "PIPELINE", pipeline)
        panpipe_labels = process_labels(panpipeconfig_json,panpipeconfig_file,panpipe_labels,pipeline)

        # rather rigid - but we foreshadow the worfflow directory here and use a param so it is useable
        wf_dir = f"<PIPELINE_DIR>/{pipeline}/<PARTICIPANT_XNAT_PROJECT>/sub-<PARTICIPANT_LABEL>/ses-<PARTICIPANT_SESSION>/{pipeline}_wf"
        panpipe_labels = updateParams(panpipe_labels,f"WORKFLOW_DIR",wf_dir)

        # We handle the <DEPENDENCY> key specially as this is a list that we need to resolve into DEPENDENCY1, DEPENDENCY2, ...DEPENDENCYN
        dependency_list = getParams(panpipe_labels,"DEPENDENCY")
        if dependency_list:
            depcount=1
            if isinstance(dependency_list,list):
                for dependency in dependency_list:
                    panpipe_labels = updateParams(panpipe_labels,f"DEPENDENCY{depcount}",dependency)
                    dependency_dir = f"<PIPELINE_DIR>/<DEPENDENCY{depcount}>/<PARTICIPANT_XNAT_PROJECT>/sub-<PARTICIPANT_LABEL>/ses-<PARTICIPANT_SESSION>/<DEPENDENCY{depcount}>_wf"
                    panpipe_labels = updateParams(panpipe_labels,f"DEPENDENCY{depcount}_DIR",dependency_dir)
                    depcount = depcount + 1
            else:
                # can refere to single DEPENDENCY also as DEPENDENCY1
                panpipe_labels = updateParams(panpipe_labels,f"DEPENDENCY1",dependency_list)
                dependency_dir1 = f"<PIPELINE_DIR>/<DEPENDENCY{depcount}>/<PARTICIPANT_XNAT_PROJECT>/sub-<PARTICIPANT_LABEL>/ses-<PARTICIPANT_SESSION>/<DEPENDENCY{depcount}>_wf"
                panpipe_labels = updateParams(panpipe_labels,f"DEPENDENCY{depcount}_DIR",dependency_dir1)
                dependency_dir = f"<PIPELINE_DIR>/<DEPENDENCY>/<PARTICIPANT_XNAT_PROJECT>/sub-<PARTICIPANT_LABEL>/ses-<PARTICIPANT_SESSION>/<DEPENDENCY>_wf"
                panpipe_labels = updateParams(panpipe_labels,f"DEPENDENCY_DIR",dependency_dir)
            
        # Now we can resolve all the references based on precedence in the pipeline section
        panpipe_labels = update_labels(panpipe_labels)

        analysis_level = getParams(panpipe_labels,"ANALYSIS_LEVEL")

        processing_environment=getParams(panpipe_labels,"PROCESSING_ENVIRONMENT") 

        if processing_environment is None  or processing_environment not in PROCESSING_OPTIONS:
            processing_environment = "slurm"

        if processing_environment == "slurm":
            analysis_node = getParams(panpipe_labels,"ANALYSIS_NODE")

            if analysis_level == "group":
                updateParams(panpipe_labels, "SLURM_TEMPLATE", getParams(panpipe_labels,"SLURM_GROUP_TEMPLATE"))
                if participant_label:
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_LABEL",participant_label)
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_PROJECT",project_list)
                    updateParams(panpipe_labels,"GROUP_SESSION_LABEL",session_list)
                else:
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_LABEL","*")
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_PROJECT","*")
                    updateParams(panpipe_labels,"GROUP_SESSION_LABEL","*")

            else:
                updateParams(panpipe_labels, "SLURM_TEMPLATE", getParams(panpipe_labels,"SLURM_PARTICIPANT_TEMPLATE"))

            if analysis_node == "gpu":
                updateParams(panpipe_labels, "SLURM_HEADER", getParams(panpipe_labels,"SLURM_GPU_HEADER"))
            else:
                updateParams(panpipe_labels, "SLURM_HEADER", getParams(panpipe_labels,"SLURM_CPU_HEADER"))

            try:
                jobid = submit_script(participant_list, participants_file, pipeline, panpipe_labels,job_ids, analysis_level,projects_list = project_list, sessions_list=session_list, sessions_file = sessions_file, LOGGER=LOGGER,panlabel=panlabel)

                job_ids[pipeline]=jobid
            except Exception as ex:
                LOGGER.info(f"problems running pipeline {pipeline}. Details: {ex}")

        elif processing_environment == "local":
            pipeline_class = getParams(panpipe_labels,"PIPELINE_CLASS")
            if pipeline_class is None:
                pipeline_class = pipeline 

            #use_pipeline_desc = getParams(panpipe_labels,"USE_PIPELINE_DESC")
            #if use_pipeline_desc is None:
            #    use_pipeline_desc = "N"

            #pipeline_desc = getParams(panpipe_labels,"PIPELINE_DESC")
            #if pipeline_desc is None or use_pipeline_desc == "N":
            #    pipeline_desc = pipeline
            #else:
            #    pipeline_desc = "".join([x if x.isalnum() else "_" for x in pipeline_desc])  

            pipeline_outdir=os.path.join(getParams(panpipe_labels,"PIPELINE_DIR"),pipeline)
            if not os.path.exists(pipeline_outdir):
                os.makedirs(pipeline_outdir,exist_ok=True)

            lock_dir=getParams(panpipe_labels,"LOCK_DIR")
            if lock_dir:
                if not os.path.exists(lock_dir):
                    os.makedirs(lock_dir,exist_ok=True)
            else:
                lock_dir=os.path.join(getParams(panpipe_labels,"PIPELINE_DIR"),"datalocks")
                if not os.path.exists(lock_dir):
                    os.makedirs(lock_dir,exist_ok=True)
                panpipe_labels = updateParams(panpipe_labels,"LOCK_DIR",lock_dir)

            bids_dir = getParams(panpipe_labels,"BIDS_DIR")

            execution_json=getParams(panpipe_labels,"NIPYPE_CONFIG")
            if execution_json is None:
                execution_json = {} 
            
            try:
                procs = getProcNums(panpipe_labels)

                if analysis_level == "participant":
                    # run serially if procs < 2
                    if procs > 1:
                        multiproc_zip = zip(participant_list,project_list,session_list)
                        parrunSingleSubject=partial(runSingleSubject, pipeline=pipeline, pipeline_class=pipeline_class, pipeline_outdir=pipeline_outdir, panpipe_labels=panpipe_labels,bids_dir=bids_dir,cred_user=cred_user,cred_password=cred_password, execution_json=execution_json,analysis_level=analysis_level)
                        with mp.Pool(procs) as pool:
                            completedlist = pool.starmap(parrunSingleSubject,multiproc_zip)
                            pool.close()
                            pool.join()
                    else:
                        for part_count in range(len(participant_list)):
                            if session_list:
                                session_label= session_list[part_count]
                            else:
                                session_label=None
                            runSingleSubject(participant_list[part_count], project_list[part_count],session_label,pipeline=pipeline, pipeline_class=pipeline_class, pipeline_outdir=pipeline_outdir, panpipe_labels=panpipe_labels,bids_dir=bids_dir,cred_user=cred_user,cred_password=cred_password, execution_json=execution_json,analysis_level=analysis_level)
                else:
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_LABEL",participant_list)
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_PROJECT",project_list)
                    updateParams(panpipe_labels,"GROUP_SESSION_LABEL",session_list)
                    runGroupSubjects(participant_list, project_list,session_list,pipeline=pipeline, pipeline_class=pipeline_class, pipeline_outdir=pipeline_outdir, panpipe_labels=panpipe_labels,bids_dir=bids_dir,cred_user=cred_user,cred_password=cred_password, execution_json=execution_json,analysis_level=analysis_level)


            except Exception as ex:
                LOGGER.info(f"problems running pipeline {pipeline}. Details: {ex}")
 
        else:
            LOGGER.info(f"processing environment {processing_environment} not currently supported. Options are {PROCESSING_OPTIONS}.")

        # clear out and restore runtime labels for next pipeline
        panpipe_labels = {}
        panpipe_labels = add_labels(runtime_labels,panpipe_labels)
       
    LOGGER.info(f"All pipelines {pipelines} successfully run/submitted.")
    LOGGER.info(f"View logs at {LOGFILE}.")
    LOGGER.debug("\n------  environment dump -----------")
    LOGGER.debug(os.environ)


# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()


__all__ = [
    "noddi",
    "qsiprep",
    "freesurfer",
    "antstransform",
    "glm_randomize_group",
    "tensor",
    "roi_mean_single",
    "roi_extract",
    "roi_mean_group",
    "parse_textdata",
    "collate_csv_single",
    "collate_csv_group",
    "basil",
    "fslanat",
    "aslprep",
    "fmriprep",
    "dummy_node",
    "atlascreate.py",
    "lst.py"
]



import re
from pathlib import Path
import os
import numpy as np
import pandas as pd
import datetime
import subprocess
import shlex
import json
import glob
import tempfile
import math
import multiprocessing as mp
import logging
import nibabel as nib
import shutil
from panpipelines.utils.transformer import *
import sys
from nipype import logging as nlogging
import fcntl
import time
from bids import BIDSLayout

UTLOGGER=nlogging.getLogger('nipype.utils')

def logger_setup(logname, loglevel):
    LOGGER = logging.getLogger(logname)
    LOGGER.setLevel(loglevel)
    return LOGGER

def logger_addstdout(LOGGER, loglevel):
    formatter = logging.Formatter('%(name)s | %(asctime)s | %(levelname)s | %(message)s')
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(loglevel)
    stdout_handler.setFormatter(formatter)
    LOGGER.addHandler(stdout_handler)

def logger_addfile(LOGGER, LOGFILE, loglevel):
    formatter = logging.Formatter('%(name)s | %(asctime)s | %(levelname)s | %(message)s')
    file_handler = logging.FileHandler(LOGFILE)
    file_handler.setLevel(loglevel)
    file_handler.setFormatter(formatter)
    LOGGER.addHandler(file_handler)

def nipype_loggers_setup(loglevel,LOGFILE,file_loglevel):
    WFLOGGER=nlogging.getLogger('nipype.workflow')
    IFLOGGER=nlogging.getLogger('nipype.interface')
    UTLOGGER=nlogging.getLogger('nipype.utils')

    formatter = logging.Formatter('%(name)s | %(asctime)s | %(levelname)s | %(message)s')

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(loglevel)
    stdout_handler.setFormatter(formatter)

    file_handler = logging.FileHandler(LOGFILE)
    file_handler.setLevel(file_loglevel)
    file_handler.setFormatter(formatter)

    if len(WFLOGGER.handlers) < 1:
        WFLOGGER.addHandler(stdout_handler)
        WFLOGGER.addHandler(file_handler)

    if len(IFLOGGER.handlers) < 1:
        IFLOGGER.addHandler(stdout_handler)
        IFLOGGER.addHandler(file_handler)

    if len(UTLOGGER.handlers) < 1:
        UTLOGGER.addHandler(stdout_handler)
        UTLOGGER.addHandler(file_handler)


def path_exists(path, parser):
    """Ensure a given path exists."""
    if path is None or not Path(path).exists():
        raise parser.error(f"Path does not exist: <{path}>.")
    return Path(path).expanduser().absolute()

def drop_sub(value):
    return re.sub(r"^sub-", "", str(value))

def drop_ses(value):
    return re.sub(r"^ses-", "", str(value))

def isTrue(arg):
    return arg is not None and (arg == 'Y' or arg == 'y' or arg == '1' or arg == 'True' or arg == 'true')

def getParams(pardict, key, update=True):
    if key is not None and pardict is not None:
        if key in pardict:
            if update:
                updateParams(pardict,key,pardict[key])
                return pardict[key] 
            else:
                return pardict[key]
    return None

def insertParams(pardict, key, value, postpone=False):
    """ Insert key:value pair into dictionary if it does not already exist.

    Parameters:
    -----------
    pardict  : The dictionary reference
    key      : Key to insert
    value    : value to associate with key
    postpone : bool , default False; If True then do not substitute variables enclosed by <>
    
    Return:
    -------
    Reference to Dictionary is returned

    """
    if key is not None and pardict is not None and value is not None and key not in pardict.keys():
        if not postpone:
            pardict[key]=substitute_labels(value,pardict)
        else:
            pardict[key]=value
    return pardict 

def updateParams(pardict, key, value, postpone=False):
    if key is not None and pardict is not None and value is not None:
        if not postpone:
            pardict[key]=substitute_labels(value,pardict)
        else:
            pardict[key]=value
    return pardict 

def removeParam(pardict,key):
    if key is not None and pardict is not None and key in pardict.keys():
        pardict.pop(key)

    return pardict

def export_labels(panpipe_labels,export_file):
    with open(export_file,"w") as outfile:
        json.dump(panpipe_labels,outfile,indent=2)


def substitute_labels(expression,panpipe_labels,exceptions=[]):
    if isinstance(expression,str):
        braced_vars = re.findall(r'\<.*?\>',expression)
        for braced_var in braced_vars:
            unbraced_var = braced_var.replace('<','').replace('>','')
            lookup_var = getParams(panpipe_labels,unbraced_var)
            if isinstance(lookup_var,str) and lookup_var is not None and unbraced_var not in exceptions:
                expression = expression.replace(braced_var,lookup_var)            
    return expression

def remove_labels(labels_dict, config_json,pipeline):  
    # Process Labels

    try:
        LABELKEY=pipeline
        LABELS=config_json[LABELKEY]
    except KeyError:
        print("'{}' not defined in labels dictionary".format(LABELKEY))
        return labels_dict

    for itemkey,itemvalue in LABELS.items():
            removeParam(labels_dict,itemkey)

    return labels_dict

def add_labels(config_dict,labels_dict,postpone=False):  
    # Process Labels
    for itemkey,itemvalue in config_dict.items():
            updateParams(labels_dict,itemkey,itemvalue,postpone=postpone)

    return labels_dict

def process_labels(config_json,config_file,labels_dict,pipeline=None,uselabel=True,insert=False,postpone=False):  
    # Process Labels

    try:
        if pipeline is None or pipeline=="":
            LABELKEY="all_pipelines"
        else:
            LABELKEY=pipeline

        if uselabel:
            LABELS=config_json[LABELKEY]
        else:
            LABELS=config_json
    except KeyError:
        print("'{}' not defined in config file {}".format(LABELKEY,str(config_file)))
        return labels_dict

    for itemkey,itemvalue in LABELS.items():
        if insert:
            insertParams(labels_dict,itemkey,itemvalue,postpone=postpone)
        else:
            updateParams(labels_dict,itemkey,itemvalue,postpone=postpone)

    return labels_dict
    

def update_labels(labels_dict):  
    # update Labels
    varkeys = list(labels_dict.keys())
    ordered_varkeys = order_varkeys(varkeys,labels_dict,varkeys.copy())

    for itemkey in ordered_varkeys:
        itemvalue = labels_dict[itemkey]
        labels_dict[itemkey]=substitute_labels(itemvalue,labels_dict)

    return labels_dict

def order_varkeys(varkeys, labels_dict, orderedkeys):
    for key in varkeys:
        if key in labels_dict.keys():
            keyindex = orderedkeys.index(key)
            if isinstance(labels_dict[key],str):
                braced_vars = re.findall(r'\<.*?\>',labels_dict[key])
                if braced_vars:
                    for braced_var in braced_vars:
                        unbraced_var = braced_var.replace('<','').replace('>','')
                        if unbraced_var in orderedkeys:
                            varkeyindex = orderedkeys.index(unbraced_var)
                            if varkeyindex > keyindex:
                                orderedkeys.insert(keyindex,unbraced_var)
                                orderedkeys.pop(varkeyindex+1)
        
    return orderedkeys


def process_dict(config_dict,labels):  
    for configkey,configvalue in labels.items():
        try:
            CONFIG=config_dict[configkey]
            for itemkey,itemvalue in configvalue.items():
                try:
                    CONFIG[itemkey]=itemvalue
                except KeyError:
                    print("{} not defined in config dict using key {}".format(str(itemkey),configkey))
                    continue

            config_dict[configkey]=CONFIG

        except KeyError:
            print("{} not defined in config dict".format(str(configkey)))
            continue

    return config_dict


def process_fsl_glm(panpipe_labels):
    fsldesign_text = getParams(panpipe_labels,"TEXT_FSL_DESIGN")
    fslcontrast_text = getParams(panpipe_labels,"TEXT_FSL_CONTRAST")
    fslftest_text = getParams(panpipe_labels,"TEXT_FSL_FTEST")

    command_base, container = getContainer(panpipe_labels,nodename="process_fsl_glm",SPECIFIC="NEURO_CONTAINER")

    if os.path.exists(fsldesign_text):
        designdir=os.path.dirname(fsldesign_text)
        outputdir=os.path.join(designdir,"fslglm")

        df = pd.read_table(fsldesign_text,sep=",",header=None)

        if not os.path.isdir(outputdir):
            os.makedirs(outputdir)

        newfsldesign_text=os.path.join(designdir,os.path.basename(fsldesign_text).split(".")[0] + '.textmat')
        newfsldesign=os.path.join(outputdir,os.path.basename(fsldesign_text).split(".")[0] + '.mat')

        df.pop(0)
        df.to_csv(newfsldesign_text,sep=" ",header=False, index=False)

        command=f"{command_base}"\
            " Text2Vest " + newfsldesign_text + " " + newfsldesign
        evaluated_command=substitute_labels(command, panpipe_labels)
        results = runCommand(evaluated_command)
        panpipe_labels = updateParams(panpipe_labels,"FSL_DESIGN",newfsldesign)


        if fslcontrast_text is not None and os.path.exists(fslcontrast_text):
            newfslcontrast=os.path.join(outputdir,os.path.basename(fslcontrast_text).split(".")[0] + '.con')
            command=f"{command_base}"\
            " Text2Vest " + fslcontrast_text + " " + newfslcontrast
            evaluated_command=substitute_labels(command, panpipe_labels)
            results = runCommand(evaluated_command)
            panpipe_labels = updateParams(panpipe_labels,"FSL_CONTRAST",newfslcontrast)
        else:
            print("TEXT_FSL_CONTRAST Contrast not defined or doed not exist")


        if fslftest_text is not None and os.path.exists(fslftest_text):
            newfslftest=os.path.join(outputdir,os.path.basename(fslftest_text).split(".")[0] + '.fts')
            command=f"{command_base}"\
            " Text2Vest " + fslftest_text + " " + newfslftest
            evaluated_command=substitute_labels(command, panpipe_labels)
            results = runCommand(evaluated_command)
            panpipe_labels = updateParams(panpipe_labels,"FSL_FTEST",newfslftest)
        else:
            print("TEXT_FSL_FTEST Ftest not defined or doed not exist")

    else:
        print("TEXT_FSL_DESIGN Design not defined or doed not exist")

def save_image_to_disk(in_file,newimgdata,output_file):
    from nilearn.image import new_img_like
    import nibabel
    
    img = nibabel.load(in_file)
    img_dtype = img.header.get_data_dtype()

    data_to_save=newimgdata.astype(img_dtype)

    new_img=new_img_like(img, data_to_save,copy_header=True)
    nibabel.nifti1.save(new_img, output_file)

def get_entities(file):
    dir_name = os.path.dirname(file)
    file_name = os.path.basename(file)

    entity={}

    if file is not None and not file == "":
        acquisition = get_bidstag("acq",file)
        if acquisition is not None:
            entity["acquisition"]=acquisition.split("acq-")[1]
            file = file.replace(acquisition, "")

    if file is not None and not file == "":
        modality = get_modality(file)
        if modality is not None:
            entity["modality"]=modality
            file = file.replace(modality, "")
    
    return entity


def get_bidstag(tag,file,all_occurences=False):
    dir_name = os.path.dirname(file)
    file_name = os.path.basename(file)

    tag=list(filter(lambda x: "{}-".format(tag) in x, file_name.split("_")))
    if tag:
        if all_occurences:
            return tag
        else:
            return tag[0]
    else:
        return None

def insert_modality(tag,file):
    dir_name = os.path.dirname(file)
    file_name = os.path.basename(file)
    ind=file_name.index(".")
    file_name = file_name[:ind] + "_" + tag + file_name[ind:] 

    file = os.path.join(dir_name,file_name)  
    return file

def get_modality(file):
    dir_name = os.path.dirname(file)
    file_name = os.path.basename(file)
    noext=file_name.split(".")[0]
    modality=noext.split("_")[-1] 
    return modality


def insert_bidstag(tag,file,overwrite=True):

    dir_name = os.path.dirname(file)
    file_name = os.path.basename(file) 
    file_parts=file_name.split("_")
    
    if len(file_parts) < 2:
        file_name = insert_modality("modality",file_name)
        file_parts=file_name.split("_")


    if not tag in file_parts:
        tag_stump=tag.split("-")
        if tag_stump:
            test_tag=get_bidstag(tag_stump[0],file_name)
            if test_tag is None or overwrite:
                if tag_stump[0] == "sub":
                    if test_tag is not None:
                        subindex=file_parts.index(test_tag)
                        if subindex != 0:
                            file_parts.pop(subindex)
                            file_parts.insert(0,tag)
                        else:
                            file_parts[0]=tag
                    else:
                        file_parts.insert(0,tag)

                elif tag_stump[0] == "ses":
                    if "sub" in file_parts[0]:
                        if test_tag is not None:
                            subindex=file_parts.index(test_tag)
                            if subindex != 1:
                                file_parts.pop(subindex)
                                file_parts.insert(1,tag)
                            else:
                                file_parts[1]=tag
                        else:
                            file_parts.insert(1,tag)
                    else:
                        print("BIDS malformed - 1st part is not sub- but {}".format(file_parts[0]))
                else:
                    if "ses" in file_parts[1]:
                        if test_tag is not None:
                            subindex=file_parts.index(test_tag)
                            if subindex != 2:
                                file_parts.pop(subindex)
                                file_parts.insert(2,tag)
                            else:
                                file_parts[2]=tag
                        else:
                            file_parts.insert(2,tag)
                    else:
                        file_parts.insert(1,tag)

            else:
                print("Tag type {} exists. Select Override True to replace with {}".format(test_tag,tag))

        else:
            print("Tag {} is malformed. '-' is missing with value".format(tag))

    else:
        print("Tag {} already in {}".format(tag,file))

    file_name="_".join(file_parts)
    file = os.path.join(dir_name,file_name)
    return file

def acquire_lock(lock_path):
    
    lock_file = open(lock_path,"w")

    try:
        fcntl.lockf(lock_file,fcntl.LOCK_EX | fcntl.LOCK_NB)
        return lock_file
    except IOError:
        UTLOGGER.debug(f"Could not acquire lock at {lock_path}. Waiting...")
        return None

def release_lock(lock_file):
    try:
        if lock_file:
            lock_file.close()
    except Exception as e:
        UTLOGGER.debug(f"problem closing lockfile {lock_file}.\n{e}")

LOCK_SUFFIX=".lock"
def getSubjectBids(labels_dict,bids_dir,participant_label,xnat_project,user,password):

    lock_path = os.path.join(getParams(labels_dict,"LOCK_DIR"),participant_label + LOCK_SUFFIX)
    lock_file = acquire_lock(lock_path)

    xnat_host = getParams(labels_dict,"XNAT_HOST")

    count = 0
    # 6 minutes timeout - this should be enough!
    TIMEOUT=360
    while not lock_file:
        time.sleep(1)
        count = count + 1
        lock_file = acquire_lock(lock_path)
        # prevent indefinite loop; take our chance on a downstream error.
        if count >= TIMEOUT:
            break

    IN_XNAT =  getParams(labels_dict,"IN_XNAT")
    if IN_XNAT:
        IN_XNAT = isTrue(IN_XNAT)
    else:
        IN_XNAT = True

    try:

        FORCEDOWNLOAD=False
        force_download = getParams(labels_dict,"FORCE_BIDS_DOWNLOAD")
        if force_download and isTrue(force_download):
            FORCEDOWNLOAD=True

        if not os.path.isdir(os.path.join(bids_dir,"sub-"+participant_label)) or FORCEDOWNLOAD:
            UTLOGGER.info(f"BIDS folder for {participant_label} not present.")
            if IN_XNAT:
                command_base, container = getContainer(labels_dict,nodename="process_fsl_glm",SPECIFIC="XNATDOWNLOAD_CONTAINER")

                UTLOGGER.info("Downloading started from XNAT.")
                getSubjectSessionsXNAT(bids_dir,participant_label,"BIDS-AACAZ",xnat_project,xnat_host,user,password)
            else:
                UTLOGGER.info(f"IN_XNAT set to {IN_XNAT}. Do not have a means of obtaining data for {participant_label}. Please add this subject's data to {bids_dir}")

        else:
            print(f"BIDS folder for {participant_label} already present. No need to retrieve")

    finally:
        release_lock(lock_file)
        try:
            os.remove(lock_path)
        except Exception as e:
            pass

def getSubjectSessionsXNAT(bids_dir,subject_label,resource_label,project,host,user,password):

    import xnat
    with xnat.connect(server=host,user=user, password=password) as connection:
        project = connection.projects[project]
        subject = project.subjects[subject_label]
        experiments = subject.experiments
        expcount = len(experiments)
        for expnum in range(expcount):
            experiment = experiments[expnum]
            if resource_label in experiment.resources:
                resource = experiment.resources[resource_label]
                tmpdir = tempfile.mkstemp()[1] + "_" + subject_label
                if not os.path.exists(tmpdir):
                    os.makedirs(tmpdir,exist_ok=True)
                tmpzip = os.path.join(tmpdir,subject_label + ".zip")
                resource.download(tmpzip)
                shutil.unpack_archive(tmpzip,tmpdir)
                bidspath = glob.glob(os.path.join(tmpdir,f"*/resources/{resource_label}/files/sub*"))
                dataset_desc = glob.glob(os.path.join(tmpdir,f"*/resources/{resource_label}/files/dataset_description.json"))
                if bidspath:
                    bidspath = bidspath[0]
                    bidssubject = bidspath.split("/")[-1]
                    shutil.copytree(bidspath,os.path.join(bids_dir,bidssubject), dirs_exist_ok=True)
                    if dataset_desc:
                        target_dataset_desc =os.path.join(bids_dir,"dataset_description.json")
                        if not os.path.exists(target_dataset_desc):
                            shutil.copy(dataset_desc[0],target_dataset_desc)
                    shutil.rmtree(tmpdir)
                else:
                    UTLOGGER.info(f"Problem dowbloading data for {subject_label} using {resource_label} in {experiment.label}")



def get_freesurfer_subregionstats(stats_file, prefix="",participant_label="",session_label=""):
    if not prefix is None and not prefix =="":
        prefix=prefix+"_"
    else:
        prefix =""

    with open(stats_file,"r") as in_file:
        lines = in_file.readlines()

    tab_col_index=-2
    tab_val_index=-1
    get_col_index=0


    table_columns = [x.split()[tab_col_index] for x in lines]
    table_columns = [x.replace('\n','') for x in table_columns]
    table_columns = table_columns[get_col_index:]
    table_columns =[ prefix + x  + "_Volume" for x in table_columns]
    table_values = [x.split()[-tab_val_index] for x in lines]
    table_values = table_values[get_col_index:]

    if len(table_columns) > 0 and len(table_values) > 0 and len(table_columns) == len(table_values):
        cum_df = pd.DataFrame([table_values])
        cum_df.columns = table_columns
        if session_label is not None and not session_label == "":
            cum_df.insert(0,"session_id",["ses-"+session_label])
        if participant_label is not None and not participant_label == "":
            cum_df.insert(0,"subject_id",["sub-"+participant_label])
        return cum_df
    else:
        return None

def get_freesurfer_hippostats(stats_file, prefix="",participant_label="", session_label=""):
    if not prefix is None and not prefix =="":
        prefix=prefix+"_"
    else:
        prefix =""

    with open(stats_file,"r") as in_file:
        lines = in_file.readlines()

    table_columns = [x.split(" ")[-1] for x in lines]
    table_columns = [x.replace('\n','') for x in table_columns]
    table_columns = table_columns[1:]
    table_columns =[ prefix + x  + "_Volume" for x in table_columns]
    table_values = [x.split()[-2] for x in lines]
    table_values = table_values[1:]

    if len(table_columns) > 0 and len(table_values) > 0 and len(table_columns) == len(table_values):
        cum_df = pd.DataFrame([table_values])
        cum_df.columns = table_columns
        if session_label is not None and not session_label == "":
            cum_df.insert(0,"session_id",["ses-"+session_label])
        if participant_label is not None and not participant_label == "":
            cum_df.insert(0,"subject_id",["sub-"+participant_label])
        return cum_df
    else:
        return None



def get_freesurfer_genstats(stats_file,columns, prefix="",participant_label="",session_label=""):

    if not prefix is None and not prefix =="":
        prefix=prefix+"_"
    else:
        prefix =""

    column_dict={}
    column_dict["StructName"] = 1
    for column in columns:
        column_dict[column] = 1

    header_list=[]
    value_list=[]

    with open(stats_file,"r") as in_file:
        lines = in_file.readlines()
    for line in lines:
        if line.startswith("# Measure"):
            header = line.split(",")[1].strip()
            header_list.append(prefix+header)
            value = line.split(",")[-2].strip()
            value_list.append(value)
        if line.startswith("# ColHeaders"):
            colheaders=line.split()[2:]
            for itemkey, itemvalue in column_dict.items():
                column_dict[itemkey]=colheaders.index(itemkey)

        if not line.startswith("#"):
            header=line.split()[column_dict["StructName"]]
            for itemkey, itemvalue in column_dict.items():
                if not itemkey == "StructName":
                    header_list.append(prefix + header + "_{}".format(itemkey))
                    value=line.split()[itemvalue]
                    value_list.append(value)
    
    if len(header_list) > 0 and len(value_list) > 0 and len(header_list) == len(value_list):
        cum_df = pd.DataFrame([value_list])
        cum_df.columns = header_list
        if session_label is not None and not session_label == "":
            cum_df.insert(0,"session_id",["ses-"+session_label])
        if participant_label is not None and not participant_label == "":
            cum_df.insert(0,"subject_id",["sub-"+participant_label])
        return cum_df
    else:
        return None

def isTrue(arg):
    return arg is not None and (arg == 'Y' or arg == '1' or arg == 'True' or arg == 'true')


def get_value_bytype(vartype,varstring):
    varvalue = None
    if vartype is not None and varstring is not None:
        if vartype  == 'text':
            varvalue=str(varstring)
        elif vartype  == 'float':
            varvalue=float(str(varstring))
        elif vartype  == 'integer':
            varvalue=int(str(varstring))
        elif vartype  == 'boolean':
            varvalue=isTrue(str(varstring))
        elif vartype  == 'list':
            if str(varstring).startswith("[") and str(varstring).endswith("]"):
                varvalue=eval(str(varstring))
        elif vartype  == 'dict':
            if str(varstring).startswith("{") and str(varstring).endswith("}"):
                varvalue=eval(str(varstring))
        else:
            varvalue=str(varstring)

    return varvalue

def create_array(participants, participants_file, projects_list = None, sessions_list=None, sessions_file = None, LOGGER=UTLOGGER):
    if sessions_file is not None:
        df = pd.read_table(sessions_file,sep="\t")
    else:
        df = pd.read_table(participants_file,sep="\t")

    array=[]
    if participants is not None and len(participants) > 0 and sessions_list and projects_list and sessions_file:
        for part_count in range(len(participants)):
            array_index = df.loc[(df["bids_participant_id"] == "sub-" + drop_sub(participants[part_count])) & (df["project"] == projects_list[part_count]) & (df["bids_session_id"].str.contains(sessions_list[part_count]))].index.values[0] + 1
            array.append(str(array_index))
        array.sort()
        return  ",".join(array)
    elif participants is not None and len(participants) > 0:
        for participant in participants:
            try:
                array.append(str(df[df["bids_participant_id"]== "sub" + drop_sub(participant)].index.values[0] + 1))
            except Exception as exp:
                UTLOGGER.debug(f"problem finding participant: {participant}")

        array.sort()
        return  ",".join(array)
    else:
        return "1:" + str(len(df))

def get_projectmap(participants, participants_file,session_labels=[],sessions_file = None, subject_exclusions=[]):

    if participants_file is not None:
        df = pd.read_table(participants_file,sep="\t")
    else:
        df = pd.read_table(sessions_file,sep="\t")

    if len(participants) == 1 and participants[0]=="ALL_SUBJECTS":
        participants = df["bids_participant_id"].tolist()

    # process exclusions
    participants = list(set(participants).difference(set(subject_exclusions)))

    # sessions are defined and so we will use this as priority
    project_list=[]
    participant_list=[]
    sessions_list=[]
    if sessions_file is not None and session_labels:
        sessions_df = pd.read_table(sessions_file,sep="\t")
        # participants and sessions are defined
        if participants is not None and len(participants) > 0:
            for participant in participants:
                if session_labels[0]=="ALL_SESSIONS":
                    search_df = sessions_df[(sessions_df["bids_participant_id"]=="sub-" + drop_sub(participant))]
                    ses=[drop_ses(ses) for ses in list(search_df.bids_session_id.values)]
                    sessions_list.extend(ses)
                    sub=[drop_sub(sub) for sub in list(search_df.bids_participant_id.values)]
                    participant_list.extend(sub)
                    proj=[proj for proj in list(search_df.project.values)]
                    project_list.extend(proj)
                else: 
                    for session_label in session_labels:
                        search_df = sessions_df[(sessions_df["bids_participant_id"]=="sub-" + drop_sub(participant)) & (sessions_df["bids_session_id"].str.contains(session_label))]
                        if search_df.empty:
                            UTLOGGER.info(f"No values found for {participant} and {session_label} in {sessions_file}")
                        else:
                            ses=[drop_ses(ses) for ses in list(search_df.bids_session_id.values)]
                            sessions_list.extend(ses)
                            sub=[drop_sub(sub) for sub in list(search_df.bids_participant_id.values)]
                            participant_list.extend(sub)
                            proj=[proj for proj in list(search_df.project.values)]
                            project_list.extend(proj)
        else:
            UTLOGGER.info(f"Cannot process pipelines. No participants have been specified")
    else:
        if participants is not None and len(participants) > 0:
            for participant in participants:
                project_list.append(str(df[df["bids_participant_id"]==participant].project.values[0]))
            sessions_list=[None for proj in project_list]
        else:
            UTLOGGER.info(f"Cannot process pipelines. No participants have been specified")

    return  [ participant_list, project_list,sessions_list ]


def create_script(header,template,panpipe_labels, script_file, LOGGER=UTLOGGER):
    with open(header,"r") as infile:
        headerlines=infile.readlines()

    with open(template,"r") as infile:
        templatelines = infile.readlines()

    newscript=[]
    header = "".join(headerlines)
    header = header + "\n\n"
    newscript.append(header)

    for templateline in templatelines:
        newscript.append(substitute_labels(templateline, panpipe_labels))

    with open(script_file,"w") as outfile:
        outfile.writelines(newscript)


def getDependencies(job_ids,panpipe_labels,LOGGER=UTLOGGER):

    slurm_dependency = getParams(panpipe_labels,"SLURM_DEPENDENCY")
    if not slurm_dependency:
        slurm_dependency="afterany"


    dependency_string=""
    pipeline_dependency = getParams(panpipe_labels,"DEPENDENCY")
    if pipeline_dependency is not None:
        if isinstance(pipeline_dependency,list):
            job_ids_string=""
            for pipe_dep in pipeline_dependency:
                if pipe_dep in job_ids.keys():
                    job_id = job_ids[pipe_dep]
                    if job_id is not None:
                        job_ids_string=job_ids_string + ":" + job_id
                    
            if not job_ids_string == "":
                dependency_string = f"--dependency={slurm_dependency}{job_ids_string}"               


        else:
            if pipeline_dependency in job_ids.keys():
                job_id = job_ids[pipeline_dependency]
                if job_id is not None:
                    dependency_string = f"--dependency={slurm_dependency}:{job_id}"

    return dependency_string



def submit_script(participants, participants_file, pipeline, panpipe_labels,job_ids, analysis_level,projects_list = None, sessions_list=None, sessions_file = None, LOGGER=UTLOGGER, script_dir=None, panlabel=None):
    headerfile=getParams(panpipe_labels,"SLURM_HEADER")
    templatefile=getParams(panpipe_labels,"SLURM_TEMPLATE")
    if not panlabel:
        panlabel = datetime.datetime.now().strftime("%Y%m%d_%H%M%S_%f") 

    script_base=pipeline + "_" + panlabel
   
    if script_dir is None:
        script_dir=os.path.join(getParams(panpipe_labels,"SLURM_SCRIPT_DIR"),script_base)

    if not os.path.exists(script_dir):
        os.makedirs(script_dir)
    script_file = os.path.join(script_dir,script_base + '.pbs')
    labels_file = os.path.join(script_dir,script_base + '.config')
    updateParams(panpipe_labels, "RUNTIME_CONFIG_FILE", labels_file)

    create_script(headerfile,templatefile,panpipe_labels, script_file)
    updateParams(panpipe_labels, "PIPELINE_SCRIPT", script_file)
    dependencies = getDependencies(job_ids,panpipe_labels)

    if LOGGER:
        LOGGER.info(f"Runtime Config file created: {labels_file}")
        LOGGER.info(f"Pipeline Script file created: {script_file}")
    
    if analysis_level == "participant":

        outlog =f"log-%a_{pipeline}_{panlabel}.panout"
        jobname = f"{pipeline}_pan_ss"

        array = create_array(participants, participants_file,projects_list = projects_list, sessions_list=sessions_list, sessions_file = sessions_file)
        
        command = "sbatch"\
        " --job-name " + jobname +\
        " --output " + outlog + \
        " --array=" + array + \
        " " + dependencies + \
        " " + script_file
    else:
        outlog =f"log-group_{pipeline}_{panlabel}.panout"
        jobname = f"{pipeline}_pan_grp"

        command = "sbatch"\
        " --job-name " + jobname +\
        " --output " + outlog + \
        " " + dependencies + \
        " " + script_file        
    
    evaluated_command = substitute_labels(command,panpipe_labels)
    if LOGGER:
        LOGGER.info(evaluated_command)

    updateParams(panpipe_labels, "SLURM_COMMAND", evaluated_command)
    evaluated_command_args = shlex.split(evaluated_command)

    # A bit of a hack - these labels in labels_dict prevent nupype from caching properly as they always change, so remove before saving
    removeParam(panpipe_labels, "RUNTIME_CONFIG_FILE")
    removeParam(panpipe_labels, "PIPELINE_SCRIPT")

    export_labels(panpipe_labels,labels_file)

    CURRDIR=os.getcwd()
    os.chdir(script_dir)

    results = subprocess.run(evaluated_command_args,stdout=subprocess.PIPE,stderr=subprocess.STDOUT, text=True)

    if LOGGER:
        LOGGER.info(results.stdout)
    else:
        print(results.stdout)
 
    os.chdir(CURRDIR)

    return results.stdout.split()[3]

def getGlob(globstring,multiple=False,default_result=""):
    if globstring:
        if "*" in globstring:
            glob_results = glob.glob(globstring)
            if not multiple:
                return getFirstFromList(glob_results,default_result)
            else:
                return glob_results
        else:
            return globstring
    else:
        return default_result

def getFirstFromList(itemlist,default_result=""):
    if len (itemlist) > 0:
        return(itemlist[0])
    else:
        return default_result

def newfile(outputdir=None, assocfile=None, prefix=None, suffix=None, intwix=None, extension=None, replace=False):
   
    if assocfile is None:
        assocfile = tempfile.mkstemp()[1]
        filename=os.path.basename(assocfile)
    else:
        filename=os.path.basename(os.path.abspath(assocfile))
        
    if outputdir is None:
        outputdir = os.path.dirname(os.path.abspath(assocfile))
    else:
        outputdir = os.path.abspath(outputdir)
    
    if not prefix is None:
        if not replace:
            filename=prefix + "_" + filename
        else:
            fileparts = filename.split("_")
            if len(fileparts) > 1:
                fileparts[0]=prefix
                filename = "_".join(fileparts)
            else:
                filename=prefix + "_" + filename

    if not extension is None:
        fileparts = filename.split(".")
        basename = fileparts[0]
        if not extension[0] == ".":
            extension="." + extension
        filename = basename + extension        
        
    if not suffix is None:
        fileparts = filename.split(".")
        basename = fileparts[0]
        extension = ".".join(fileparts[1:])
        if len(extension) > 0:
            if not extension[0] == ".":
                extension="." + extension
        if not replace:
            filename = basename + "_" + suffix + extension
        else:
            fileparts = basename.split("_")
            if len(fileparts) > 1:
                fileparts[-1]=suffix
                basename = "_".join(fileparts)
                filename = basename + extension
            else:
                filename = basename + "_"+ suffix + extension

    if not intwix is None:
        filename = "_".join(filename.split("_")[0:-1] + [intwix] + [filename.split("_")[-1]])


                
    return os.path.join(outputdir,filename)
        
def getTransName(fromfile, tofile):
    if fromfile is not None and tofile is not None:
        return "from-" + os.path.basename(fromfile).split(".")[0] + "_" + "to-" + os.path.basename(tofile).split(".")[0]
        
def getProcNums(panpipe_labels):

    procnum_list=[mp.cpu_count()]

    try:
        pipeline_count = int(getParams(panpipe_labels,"PIPELINE_THREADS"))
        procnum_list.append(pipeline_count)
    except ValueError as ve:
        pass
    except TypeError as te:
        pass
    

    try:
        ENV_PROCS = getParams(panpipe_labels,"ENV_PROCS")
        if ENV_PROCS is not None and ENV_PROCS in os.environ.keys():
            env_count = int(os.environ[ENV_PROCS])
            procnum_list.append(env_count)
    except ValueError as ve:
        pass
    except TypeError as te:
        pass
    
    return int(np.min(np.array(procnum_list)))


def add_atlas_roi(atlas_file, roi_in, roi_value, panpipe_labels, high_thresh=None,low_thresh=None,prob_thresh=0.5,roi_transform=None):

    workdir = os.path.join(os.path.dirname(atlas_file),'roi_temp')
    if not os.path.isdir(workdir):
        os.makedirs(workdir)

    trans_workdir = os.path.join(os.path.dirname(atlas_file),'roi_transformed')
    if not os.path.isdir(trans_workdir):
        os.makedirs(trans_workdir)

    if prob_thresh:
        PROBTHRESH=f" -thr {prob_thresh}"
    else:
        PROBTHRESH=""

    if high_thresh:
        HIGHTHRESH=f" -uthr {high_thresh}"
    else:
        HIGHTHRESH = ""

    if low_thresh:
        LOWTHRESH=f" -thr {low_thresh}"
    else:
        LOWTHRESH=""

    # store roi in work dir for
    command_base, container = getContainer(panpipe_labels,nodename="add_atlas_roi",SPECIFIC="NEURO_CONTAINER")
    new_roi=newfile(trans_workdir, roi_in, suffix="desc-label")
    command = f"{command_base} fslmaths"\
        f"  {roi_in}" +\
        PROBTHRESH +\
        " -bin "\
        f" -mul {roi_value}" \
        f" {new_roi}" 
    
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)

    from panpipelines.nodes.antstransform import antstransform_proc

    roi_transform_ref = getParams(panpipe_labels,"ROI_TRANSFORM_REF")
    if roi_transform:
        CURRDIR = os.getcwd()
        os.chdir(trans_workdir)
        results = antstransform_proc(panpipe_labels, new_roi,roi_transform, roi_transform_ref)
        new_roi_transformed = results['out_file']
        os.chdir(CURRDIR)
    else:
        new_roi_transformed = newfile(trans_workdir, new_roi)
        shutil.move(new_roi, new_roi_transformed)

    # make output file into int. This was initially don in ANTS above but had issues with rois that had value of 1
    command = f"{command_base} fslmaths"\
            f"  {new_roi_transformed}"\
            f" {new_roi_transformed}" \
             " -odt int"
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)

    if os.path.exists(atlas_file):
        command = f"{command_base} fslmaths"\
            f"  {new_roi_transformed}"\
            f" -add {atlas_file}" +\
            LOWTHRESH +\
            HIGHTHRESH +\
            f" {atlas_file}"
    else:
        command = f"{command_base} fslmaths"\
            f" {new_roi_transformed}"\
            f" {atlas_file}" 
    
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)


def create_3d_atlas_from_rois(atlas_file, roi_list,panpipe_labels, roi_values=None,prob_thresh=0.5,explode3d=True):

    roi_list = expand_rois(roi_list,os.path.dirname(atlas_file),panpipe_labels,explode3d=explode3d)
    roi_transform_mat = getParams(panpipe_labels,"ROI_TRANSFORM_MAT")
    panpipe_labels=updateParams(panpipe_labels,"ROI_TRANSFORM_REF",getParams(panpipe_labels,"NEWATLAS_TRANSFORM_REF"))

    numrois=len(roi_list)
    if roi_values is None:
        roi_values=range(1,numrois+1)

    # create rois
    for roi_num in range(numrois):
        roi = roi_list[roi_num]
        roi_transform=None
        if roi_transform_mat and roi_num < len(roi_transform_mat):
            roi_transform = roi_transform_mat[roi_num]
        roi_value = roi_values[roi_num]
        add_atlas_roi(atlas_file, roi, roi_value, panpipe_labels,high_thresh=roi_value,prob_thresh=prob_thresh,roi_transform=roi_transform)

    return atlas_file

def merge_atlas_roi(atlas_file, roi_list, panpipe_labels, high_thresh=None,low_thresh=None):

    workdir = os.path.join(os.path.dirname(atlas_file),'roi_temp')
    if not os.path.isdir(workdir):
        os.makedirs(workdir)

    trans_workdir = os.path.join(os.path.dirname(atlas_file),'roi_transformed')
    if not os.path.isdir(trans_workdir):
        os.makedirs(trans_workdir)

    if high_thresh:
        HIGHTHRESH=f" -uthr {high_thresh}"
    else:
        HIGHTHRESH = ""

    if low_thresh:
        LOWTHRESH=f" -thr {low_thresh}"
    else:
        LOWTHRESH=""

    command_base, container = getContainer(panpipe_labels,nodename="merge_atlas_roi",SPECIFIC="NEURO_CONTAINER")

    roicount=0
    numrois=len(roi_list)
    roi_files=[]
    for roi_in in roi_list:       
        # store roi in work dir\
        roicount=roicount+1
        new_roi=newfile(workdir, roi_in,suffix="desc-bin")
        command = f"{command_base} fslmaths"\
            f"  {roi_in}" +\
            LOWTHRESH +\
            HIGHTHRESH +\
            " -bin "\
            f" {new_roi}"
        evaluated_command=substitute_labels(command,panpipe_labels)
        results = runCommand(evaluated_command)
        roi_files.append(new_roi)

    from panpipelines.nodes.antstransform import antstransform_proc
    roi_files_transformed=[]
    roi_transform_mat = getParams(panpipe_labels,"ROI_TRANSFORM_MAT")
    roi_transform_ref = getParams(panpipe_labels,"ROI_TRANSFORM_REF")
    for roi_num in range(len(roi_files)):
        if roi_transform_mat and roi_num < len(roi_transform_mat):
            roi_file=roi_files[roi_num]
            roi_transform = roi_transform_mat[roi_num]
            CURRDIR=os.getcwd()
            os.chdir(trans_workdir)
            results = antstransform_proc(panpipe_labels, roi_file,roi_transform, roi_transform_ref)
            os.chdir(CURRDIR)
            new_roi_transform_file=results['out_file']
        else:
            new_roi_transform_file= newfile(trans_workdir, new_roi)
            shutil.move(roi_files[roi_num], new_roi_transform_file)
        # make output file into int. This was initially don in ANTS above but had issues with rois that had value of 1
        command = f"{command_base} fslmaths"\
                f"  {new_roi_transform_file}"\
                f" {new_roi_transform_file}" \
                " -odt int"
        evaluated_command=substitute_labels(command,panpipe_labels)
        results = runCommand(evaluated_command)
        roi_files_transformed.append(new_roi_transform_file)
        
    if roi_files_transformed:
        roi_string=" ".join(roi_files_transformed)
        command = f"{command_base} fslmerge"\
            " -t" \
            f" {atlas_file}" +\
            " " + roi_string

        evaluated_command=substitute_labels(command,panpipe_labels)
        results = runCommand(evaluated_command)

    return atlas_file

def create_4d_atlas_from_rois(atlas_file, roi_list,panpipe_labels, roi_values=None, high_thresh=None, low_thresh=0.5,explode3d=True):

    panpipe_labels=updateParams(panpipe_labels,"ROI_TRANSFORM_REF",getParams(panpipe_labels,"NEWATLAS_TRANSFORM_REF"))
    roi_list = expand_rois(roi_list,os.path.dirname(atlas_file),panpipe_labels,explode3d=explode3d)
    merge_atlas_roi(atlas_file, roi_list, panpipe_labels,high_thresh=high_thresh, low_thresh=low_thresh)

    return atlas_file


def expand_rois(roi_list, out_dir, panpipe_labels,explode3d=True):
    """ Given a list of roi images with associated transforms, attempt to reconcile a mix of 3D and 4D images in the list such that
        each element in the list is a 3D roi which can be eventually merged into 1 single 3D atlas or 1 single 4D atlas.
    
    Parameters:
    -----------
    roi_list : The list of rois
    out_dir  : Output directory to store temporary files
    panpipe_labels    : Dictionary of config labels. This should contain NEURO_CONTAINER reference to singularity image
    explode3D : bool , default True. Set this to true when you want to treat each roi as a contiguous uniform label regardless of if it is not.
    
    return:
    -------
    Reference to Dictionary is returned
    """
    workdir = os.path.join(out_dir,'roi_list_temp')
    if not os.path.isdir(workdir):
        os.makedirs(workdir)

    roi_transform_list=[]
    newatlas_transform = getParams(panpipe_labels,"NEWATLAS_TRANSFORM_MAT")
    new_roi_list = []
    roi_count = 0

    command_base, container = getContainer(panpipe_labels,nodename="expand_rois",SPECIFIC="NEURO_CONTAINER")
    for roi in roi_list:

        # using fsl for manipulations, so convert freesurfer files to nifti
        if Path(roi).suffix == ".mgz":

            mgzdir = os.path.join(out_dir,'roi_mgz_temp')
            if not os.path.isdir(mgzdir):
                os.makedirs(mgzdir)
            fs_command_base, fscontainer = getContainer(panpipe_labels,nodename="convMGZ2NII",SPECIFIC="FREESURFER_CONTAINER")
            roi_nii = newfile(mgzdir,roi,extension=".nii.gz")
            convMGZ2NII(roi, roi_nii, fs_command_base)
            roi = roi_nii

        roi_transform = None
        if newatlas_transform and roi_count < len(newatlas_transform):
            roi_transform = newatlas_transform[roi_count]

        roi_count = roi_count + 1
        roi_img  = nib.load(roi)
        roi_shape = roi_img.header.get_data_shape()
        if len(roi_shape) > 3:
            for vol in range(len(roi_shape[3])):
                sub_roi_num=vol + 1
                new_roi=newfile(workdir,roi, prefix=f"{roi_count:0>5}_{sub_roi_num:0>5}",suffix="desc-roi")
                command = f"{command_base} fslroi"\
                    f"  {roi}" \
                    f" {new_roi}" \
                    f" {vol}" \
                    " 1" 
                evaluated_command=substitute_labels(command,panpipe_labels)
                results = runCommand(evaluated_command)
                new_roi_list.append(new_roi)
                if roi_transform:
                    roi_transform_list.append(roi_transform)

        else:
            if explode3d:
                atlas_dict,atlas_list=get_avail_labels(roi)
                for thresh in atlas_list:
                    new_roi=newfile(workdir,roi, prefix=f"{roi_count:0>5}_{thresh:0>5}",suffix="desc-roi")
                    command = f"{command_base} fslmaths"\
                        f" {roi}" \
                        f" -thr {thresh}" \
                        f" -uthr {thresh}" \
                        " -bin "\
                        f" {new_roi}"
                    evaluated_command=substitute_labels(command,panpipe_labels)
                    results = runCommand(evaluated_command)
                    new_roi_list.append(new_roi)
                    if roi_transform:
                        roi_transform_list.append(roi_transform)
          
            else:
                new_roi=newfile(workdir,roi, prefix=f"{roi_count:0>5}",suffix="desc-roi")
                command = "cp"\
                " " + roi +\
                " " + new_roi
                evaluated_command=substitute_labels(command,panpipe_labels)
                results = runCommand(evaluated_command)
                new_roi_list.append(new_roi)
                if roi_transform:
                        roi_transform_list.append(roi_transform)

    if roi_transform_list:
        updateParams(panpipe_labels,"ROI_TRANSFORM_MAT",roi_transform_list)
    
    return new_roi_list

def get_avail_labels(atlas_file):
    atlas_dict={}
    atlas_img = nib.load(atlas_file)
    atlas_data = atlas_img.get_fdata()
    max_roi_num = int(np.max(atlas_data))
    for roi_num in range(1,max_roi_num+1):
        num_voxels=np.count_nonzero(atlas_data == roi_num)
        if num_voxels:
            atlas_dict[roi_num]=num_voxels

    atlas_index_list=[]
    atlas_key_list=list(atlas_dict.keys())
    atlas_key_list.sort()
    for atlas_key in atlas_key_list:
        atlas_index_list.append(atlas_key)
    
    return atlas_dict,atlas_index_list


def get_freesurferatlas_index_mode(atlas_file,lutfile,atlas_index,atlas_index_mode=None):
    
    if atlas_index_mode and "freesurf_tsv_general" in atlas_index_mode:
        atlas_index_type="tsv"
        if "contig" in atlas_index_mode:
            atlas_index_type = "tsv_contig"
        atlas_dict,atlas_index_out=get_freesurferatlas_index(atlas_file,lutfile,atlas_index,avail_only=True,use_atlas_max=True,atlas_index_type=atlas_index_type)
    elif atlas_index_mode and "freesurf_general" in atlas_index_mode:
        atlas_dict,atlas_index_out=get_freesurferatlas_index(atlas_file,lutfile,atlas_index,avail_only=False,use_atlas_max=True)
    elif atlas_index_mode and "freesurf_contig_general" in atlas_index_mode:
        atlas_dict,atlas_index_out=get_freesurferatlas_index(atlas_file,lutfile,atlas_index,avail_only=True,use_atlas_max=True)
    elif atlas_index_mode and "hcpmmp1aseg_tsv" in atlas_index_mode:
        atlas_index_type="tsv"
        if "contig" in atlas_index_mode:
            atlas_index_type = "tsv_contig"
        atlas_dict,atlas_index_out=get_freesurferatlas_index(atlas_file,lutfile,atlas_index,avail_only=False,use_atlas_max=False,atlas_index_type=atlas_index_type)
    elif atlas_index_mode and "hcpmmp1aseg" in atlas_index_mode:
        atlas_dict,atlas_index_out=get_freesurferatlas_index(atlas_file,lutfile,atlas_index,avail_only=False,use_atlas_max=False)
    else:
        atlas_dict,atlas_index_out=get_freesurferatlas_index(atlas_file,lutfile,atlas_index,atlas_index_mode=atlas_index_mode)

    return atlas_dict,atlas_index_out


def get_freesurferatlas_index(atlas_file,lutfile,atlas_index,avail_only=False,use_atlas_max=True, atlas_index_type=None):
    with open(lutfile,'r') as infile:
        lutlines=infile.readlines()

    lut_dict={}
    lut_max = 0
    for lut_line in lutlines:
        lut_list = lut_line.split()
        if lut_list and lut_list[0].isdigit():           
            lut_roinum = int(lut_list[0])
            lut_max=max(lut_max,lut_roinum)
            lut_roiname = lut_list[1]
            lut_rgba = ",".join(lut_list[2:])
            lut_dict[lut_roinum] = {}
            lut_dict[lut_roinum]["LabelName"]=lut_roiname
            lut_dict[lut_roinum]["RGBA"]=lut_rgba

    atlas_dict={}
    atlas_img = nib.load(atlas_file)
    atlas_data = atlas_img.get_fdata()
    if use_atlas_max:
        max_roi_num = int(np.max(atlas_data))
    else:
        max_roi_num = lut_max 
    for roi_num in range(1,max_roi_num+1):
        num_voxels=np.count_nonzero(atlas_data == roi_num)
        if avail_only:
            if num_voxels:
                atlas_dict[roi_num]=lut_dict[roi_num]
                atlas_dict[roi_num]["Voxels"]=num_voxels
        else:
            if roi_num in lut_dict.keys():
                atlas_dict[roi_num]=lut_dict[roi_num]
            else:
                atlas_dict[roi_num]={}
                atlas_dict[roi_num]["LabelName"]=f"UNK_{roi_num}"
                atlas_dict[roi_num]["RGBA"]="0,0,0,0"
            atlas_dict[roi_num]["Voxels"]=num_voxels

    atlas_index_list=[]
    atlas_key_list=list(atlas_dict.keys())
    atlas_key_list.sort()

    if atlas_index_type and "tsv" in atlas_index_type:
        atlas_index_list.append("index\tlabel")
        contignum=1
        for atlas_key in atlas_key_list:
            if "contig" in atlas_index_type:
                atlas_index_list.append(str(contignum) + "\t" + atlas_dict[atlas_key]["LabelName"])
            else:
                atlas_index_list.append(str(atlas_key) + "\t" + atlas_dict[atlas_key]["LabelName"])
            contignum=contignum+1
    else:
        for atlas_key in atlas_key_list:
            atlas_index_list.append(atlas_dict[atlas_key]["LabelName"])
        
    atlas_index_out = "\n".join(atlas_index_list)
    if atlas_index:
        with open(atlas_index,"w") as outfile:
            outfile.write(atlas_index_out)

    return atlas_dict,atlas_index_out

            
def create_3d_hcpmmp1_aseg(atlas_file,roi_list,panpipe_labels):
    out_dir = os.path.dirname(atlas_file)
    workdir = os.path.join(out_dir,"hcpmmp1_workdir")
    if not os.path.isdir(workdir):
        os.makedirs(workdir)

    command_base, container = getContainer(panpipe_labels,nodename="create_3d_hcppmmp1_aseg",SPECIFIC="NEURO_CONTAINER")

    atlas_dir = getParams(panpipe_labels,"ATLAS_DIR")
    SUB=f"sub-{getParams(panpipe_labels,'PARTICIPANT_LABEL')}"
    freesurfer_dir = getParams(panpipe_labels,"FREESURFER_DIR")
    freesurfer_home = getParams(panpipe_labels,"FREESURFER_HOME")
    if not freesurfer_home:
        freesurfer_home = "/opt/freesurfer"

    lh_hcpannot = os.path.join(atlas_dir, "lh.HCP-MMP1.annot")
    lh_hcpannot_trg = os.path.join(freesurfer_dir,SUB,"label","lh.HCP-MMP1.annot")
    rh_hcpannot = os.path.join(atlas_dir, "rh.HCP-MMP1.annot")
    rh_hcpannot_trg = os.path.join(freesurfer_dir,SUB,"label","rh.HCP-MMP1.annot")

    command = f"{command_base} ln"\
        "  -s" \
        f" {os.path.join(freesurfer_home,'subjects','fsaverage')}" \
        f" {os.path.join(freesurfer_dir,'fsaverage')}" 
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)

    command = f"{command_base} ln"\
        "  -s" \
        f" {os.path.join(freesurfer_home,'subjects','lh.EC_average')}" \
        f" {os.path.join(freesurfer_dir,'lh.EC_average')}" 
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)

    command = f"{command_base} ln"\
        "  -s" \
        f" {os.path.join(freesurfer_home,'subjects','rh.EC_average')}" \
        f" {os.path.join(freesurfer_dir,'rh.EC_average')}" 
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)

    os.environ["SINGULARITYENV_SUBJECTS_DIR"]=freesurfer_dir

    command = f"{command_base} mri_surf2surf"\
        " --srcsubject fsaverage" \
        f" --trgsubject {SUB}" \
        " --hemi lh" \
        f" --sval-annot {lh_hcpannot}" \
        f" --tval {lh_hcpannot_trg}"
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)

    command = f"{command_base}  mri_surf2surf"\
        " --srcsubject fsaverage" \
        f" --trgsubject {SUB}" \
        " --hemi rh" \
        f" --sval-annot {rh_hcpannot}" \
        f" --tval {rh_hcpannot_trg}"
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)

    atlas_space_fs = newfile(workdir, atlas_file, suffix="desc-hcpmmp1_space-fs")
    command = f"{command_base} mri_aparc2aseg"\
        f"  --s {SUB}" \
        "  --old-ribbon" \
        " --annot HCP-MMP1" \
        f" --o {atlas_space_fs}" 
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)

    atlas_space_T1w = newfile(workdir, atlas_file, suffix="desc-hcpmmp1_space-T1w",extension=".mgz")
    rawavg=os.path.join(freesurfer_dir,SUB,"mri","rawavg.mgz")
    command = f"{command_base} mri_label2vol"\
        f"  --seg {atlas_space_fs}" \
        f"  --temp {rawavg}" \
        f"  --o {atlas_space_T1w}" \
        f"  --regheader {atlas_space_fs}" 
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)

    atlas_space_T1w_nii = newfile(workdir, atlas_space_T1w,suffix="desc-unordered",extension=".nii.gz")
    fs_command_base, fscontainer = getContainer(panpipe_labels,nodename="convMGZ2NII",SPECIFIC="FREESURFER_CONTAINER")
    convMGZ2NII(atlas_space_T1w,atlas_space_T1w_nii, fs_command_base)


    from panpipelines.nodes.antstransform import antstransform_proc

    panpipe_labels= updateParams(panpipe_labels,"COST_FUNCTION","NearestNeighbor")

    atlas_transform_mat = getParams(panpipe_labels,"NEWATLAS_TRANSFORM_MAT")
    
    if not atlas_transform_mat:
        atlas_transform_mat = getParams(panpipe_labels,"ATLAS_TRANSFORM_MAT")

    atlas_transform_ref = getParams(panpipe_labels,"NEWATLAS_TRANSFORM_REF")
    if not atlas_transform_ref:
        atlas_transform_ref = getParams(panpipe_labels,"ATLAS_TRANSFORM_REF")

    if atlas_transform_mat:
        CURRDIR=os.getcwd()
        os.chdir(workdir)
        results = antstransform_proc(panpipe_labels, atlas_space_T1w_nii,atlas_transform_mat, atlas_transform_ref)
        os.chdir(CURRDIR)
        atlas_space_transform = results['out_file']
    else:
        atlas_space_transform=atlas_space_T1w_nii

    hcpmmp_original = os.path.join(atlas_dir, "hcpmmp1_original.txt")
    hcpmmp_ordered = os.path.join(atlas_dir, "hcpmmp1_ordered.txt")

    # enforce int data type at this point
    command = f"{command_base} fslmaths"\
            f"  {atlas_space_transform}"\
            f" {atlas_space_transform}" \
            " -odt int"
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)


    command = f"{command_base} labelconvert"\
        f"  {atlas_space_transform}" \
        f"  {hcpmmp_original}" \
        f"  {hcpmmp_ordered}" \
        f"  {atlas_file}" 
    evaluated_command=substitute_labels(command,panpipe_labels)
    results = runCommand(evaluated_command)

    return [f"get_freesurfer_atlas_index:{hcpmmp_ordered}"]


def getAge(birthdate,refdate=None):
    from datetime import date
    if refdate is None:
        today = date.today()
    else:
        today=refdate
    age = today.year - birthdate.year - ((today.month, today.day) < (birthdate.month, birthdate.day))
    return age

def initTemplateFlow(TEMPLATEFLOW_HOME):
    ref1=get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin2009cAsym",resolution=2,suffix="T1w",extension=[".nii.gz"])
    ref2=get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin6Asym",resolution=2,suffix="T1w",extension=[".nii.gz"])
    transform1 = get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin2009cAsym",suffix="xfm",extension=[".h5"])
    transform2 = get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin6Asym",suffix="xfm",extension=[".h5"])
    return [[ref1,transform1],[ref2,transform2]]

def getBidsTSV(host,user,password,projects,targetfolder,outputdir,demographics=True):
    import xnat
    from pydicom import dcmread
    import fnmatch

    if not os.path.isdir(outputdir):
        os.mkdir(outputdir)

    participantsTSV=os.path.join(outputdir,'participants.tsv')
    sessionsTSV=os.path.join(outputdir,'sessions.tsv')

    participant_columns=['xnat_subject_id','xnat_subject_label','bids_participant_id','project','gender', 'age','scan_date','comments']
    participant_data=[]

    session_columns=['xnat_session_id','xnat_session_label','xnat_subject_id','xnat_subject_label','bids_participant_id','bids_session_id', 'project','gender', 'age','scan_date','comments']
    session_data=[]

    scantest = os.path.join(outputdir,'scantest.dcm')

    with xnat.connect(server=host,user=user, password=password) as connection:

        if projects is None:
            projects = connection.projects

        for PROJ in projects:
            try: 
                project = connection.projects[PROJ]
                UTLOGGER.info(f"Getting participant information for Project {project.id}.\n")
                # rigmarole here as iteration acting weirdly with experiments
                experiments = project.experiments
                numexps =  len(experiments)
                for expid in range(numexps):

                    print_progress(expid + 1, numexps, prefix='Progress', suffix='Complete')
                    experiment=experiments[expid]
                    xnat_session_id =""
                    xnat_session_label = ""
                    xnat_subject_id = ""
                    xnat_subject_label = ""
                    bids_participant_id = ""
                    bids_session_id = ""
                    gender = ""
                    age = ""
                    comments=""
                    scan_date=""


                    # _MR_ hack - some reason 003_HML MR session not storing modality 
                    if experiment.modality == 'MR' or "_MR_" in experiment.label:
                        xnat_session_id = experiment.id
                        xnat_session_label = experiment.label
                        xnat_subject_id = experiment.subject.id
                        xnat_subject_label = experiment.subject.label
                        try:
                            if (experiment.resources[targetfolder]):
                                files = experiment.resources[targetfolder].files
                                numfiles = len(files)
                                for fileid in range(numfiles):
                                    file=files[fileid]
                                    if fnmatch.fnmatch(file.path,'sub-*/ses-*/*'):
                                        bids_participant_id = file.path.split('/')[0]
                                        bids_session_id = file.path.split('/')[1]
                                        break

                                if not bids_participant_id:
                                    for fileid in range(numfiles):
                                        file=files[fileid]
                                        if fnmatch.fnmatch(file.path,'sub-*/*'):
                                            bids_participant_id = file.path.split('/')[0]
                                            bids_session_id = ''
                                            break

                            if demographics:
                                scans = experiment.scans
                                for scan_index in range(len(scans)):
                                    if "DICOM" in experiment.scans[scan_index].resources.keys():
                                        experiment.scans[scan_index].resources['DICOM'].files[0].download(scantest)
                                        ds = dcmread(scantest)
                                        gender = ds.PatientSex
                                        if ds.PatientBirthDate:
                                            agedate=datetime.datetime.strptime(ds.PatientBirthDate,"%Y%m%d")
                                        else:
                                            agedate=experiment.subject.demographics.dob 
                                        scan_date = ds.StudyDate   
                                        scandate = datetime.datetime.strptime(scan_date,"%Y%m%d")
                                        if agedate is None or agedate == '':
                                            comments = "subject age missing; " + comments
                                        else:
                                            age=getAge(agedate,scandate)
                                        break

                        except Exception as e:
                            message = 'problem parsing resource : %s.' % targetfolder
                            comments="missing bids files; " + comments
                            print(message)
                            print(str(e))

                        session_data.append([xnat_session_id,xnat_session_label,xnat_subject_id,xnat_subject_label,bids_participant_id,bids_session_id, project.id, gender, age,scan_date,comments])

            except Exception as e:
                message = 'problem parsing project :  %s.' % PROJ
                print (message)
                print(str(e))

        df = pd.DataFrame(session_data, columns=session_columns)
        df.to_csv(sessionsTSV,sep="\t", index=False)

        first=df.groupby('xnat_subject_id').first()
        subs_list=df['xnat_subject_id'].to_list()
        subs_set=set(subs_list)
        first['xnat_subject_id'] = sorted(subs_set)
        cols=list(first.columns.values)
        cols.pop(cols.index('xnat_subject_id'))
        cols.pop(cols.index('xnat_session_id'))
        cols.pop(cols.index('xnat_session_label'))
        cols.pop(cols.index('bids_session_id'))
        newdf=first[['xnat_subject_id'] + cols]
        newdf.to_csv(participantsTSV,sep="\t", index=False)

        if os.path.exists(scantest):
            os.remove(scantest)

# Print iterations progress
def print_progress(iteration, total, prefix='', suffix='', decimals=1, bar_length=50):
    """
    Call in a loop to create terminal progress bar
    https://gist.github.com/aubricus/f91fb55dc6ba5557fbab06119420dd6a
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        bar_length  - Optional  : character length of bar (Int)
    """
    str_format = "{0:." + str(decimals) + "f}"
    percents = str_format.format(100 * (iteration / float(total)))
    filled_length = int(round(bar_length * iteration / float(total)))
    bar = '|' * filled_length + '-' * (bar_length - filled_length)

    sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, bar, percents, '%', suffix)),

    if iteration == total:
        sys.stdout.write('\n')
    sys.stdout.flush()

def printProgressBar (iteration, total, prefix = '', suffix = '', decimals = 1, length = 50, fill = '|', printEnd = "\r"):
    """
    Call in a loop to create terminal progress bar
    https://stackoverflow.com/questions/3173320/text-progress-bar-in-terminal-with-block-characters
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print(f'\r{prefix} |{bar}| {percent}% {suffix}', end = printEnd)
    # Print New Line on Complete
    if iteration == total: 
        print()

def arrangePipelines(jsondict,pipelines=[]):
    # build up dictionary of pipelines and dependencies
    pipeline_dict={}
    drop_pipelines=[]
    for pipeline in pipelines:
        if pipeline in jsondict.keys():
            pipeline_dict[pipeline]=[]
            if "DEPENDENCY" in jsondict[pipeline]:
                dependency = jsondict[pipeline]["DEPENDENCY"]
                if not isinstance(dependency,list):
                    dependency = [dependency]
                dep_list=[]
                for dep in dependency:
                    if dep in pipelines:
                        dep_list.append(dep)

                if dep_list:
                    pipeline_dict[pipeline]=dep_list
        else:
            UTLOGGER.info(f"{pipeline} not defined in configuration file. Please check spelling. Removing from list of pipelines")
            drop_pipelines.append(pipeline)

    # Drop pipelines that are not defined
    for droppipes in drop_pipelines:
        pipelines.remove(droppipes)

    arranged_pipelines=pipelines.copy()
    # Now sort pipelines according to order of dependencies.start with first pipeline and make sure it runs after dependency.
    for pipeline in pipelines:
        arranged_pipelines = shufflePipelines(pipeline_dict, pipeline, arranged_pipelines)

    return arranged_pipelines

def shufflePipelines(pipeline_dict,pipeline,pipelines):

    pipeindex = pipelines.index(pipeline)

    newpipeindex = pipeindex
    for pipe in pipelines:
        if pipe != pipeline:
            if pipe in pipeline_dict[pipeline]:
                newpipeindex=recursePipe([newpipeindex],pipe,pipelines,pipeline_dict)

    if newpipeindex != pipeindex:
        if newpipeindex > pipeindex:
            pipelines.insert(newpipeindex+1,pipeline)
            pipelines.pop(pipeindex)


    return pipelines

def recursePipe(recurselist,pipe,pipelines,pipeline_dict):
    recurselist.append(pipelines.index(pipe))
    for dep in pipeline_dict[pipe]:
        recursePipe(recurselist,dep,pipelines,pipeline_dict)
    return max(recurselist)


def getContainer(labels_dict,nodename="",SPECIFIC=None,CONTAINERALT="PAN_CONTAINER", LOGGER=UTLOGGER):
    if SPECIFIC:
        container_run_options = getParams(labels_dict,f"{SPECIFIC}_RUN_OPTIONS")
        container_prerun = getParams(labels_dict,f"{SPECIFIC}_PRERUN")
        container = getParams(labels_dict,f"{SPECIFIC}")
    else:
        container_run_options = None
        container_prerun = None
        container = None

    if not container_run_options:
        container_run_options = getParams(labels_dict,'CONTAINER_RUN_OPTIONS')
        if not container_run_options:
            container_run_options = ""

    if not container_prerun:
        container_prerun = getParams(labels_dict,'CONTAINER_PRERUN')
        if not container_prerun:
            container_prerun = ""

    if not container:
        container = getParams(labels_dict,'CONTAINER')
        if not container:
            container = getParams(labels_dict,f'{CONTAINERALT}')
            if not container:
                container = getParams(labels_dict,'NEURO_CONTAINER')
                if not container:
                    container=""

                    LOGGER.info(f"Container not defined for {nodename} node. Required commands should be accessible on local path for pipeline to succeed")
                    if container_run_options:
                        LOGGER.info(f"Note that '{container_run_options}' set as run options for non-existing container. This may cause the pipeline to fail.")
                    
                    if container_prerun:
                        LOGGER.info(f"Note that '{container_prerun}' set as pre-run options for non-existing container. This may cause the pipeline to fail.")

    command_base = f"{container_run_options} {container} {container_prerun}"
    LOGGER.info("Container base run command is:")
    LOGGER.info(f"{command_base}")

    return command_base, container

def runCommand(command,LOGGER=UTLOGGER,suppress="",interactive=False):
    if suppress:
        LOGGER.info(suppress)
    else:
        LOGGER.info(command)
    evaluated_command_args = shlex.split(command)
    if suppress:
        results = subprocess.run(evaluated_command_args,stdout=subprocess.PIPE,stderr=subprocess.STDOUT, text=True)
        return "<Suppressed>"
    else:
        if not interactive:
            results = subprocess.run(evaluated_command_args,stdout=subprocess.PIPE,stderr=subprocess.STDOUT, text=True)
            LOGGER.info(results.stdout)
            return results.stdout
        else:
            results = subprocess.run(evaluated_command_args)

# this is a specific function here for sdcflows fieldmaps - should move and make more general
def getAcquisition(bids_dir,participant_label,participant_session=None,suffix="asl",extension="nii.gz"):
    acq=""    
    layout = BIDSLayout(bids_dir)
    bidslist=layout.get(subject=participant_label,session=participant_session,suffix=suffix, extension=extension)
    if len(bidslist) > 0:
        bidsfile =bidslist[0]
        entities = bidsfile.get_entities()
        if "acquisition" in entities.keys():
            acq = "acq-" + entities["acquisition"]
        else:
            acq = get_bidstag("acq",bidsfile.filename)

    return acq

def getPhaseDiffSources(bids_dir,participant_label,participant_session=None,datatype="fmap",suffix=['phase1','phase2','magnitude1','magnitude2','phasediff','magnitude'],extension="nii.gz"):   
    layout = BIDSLayout(bids_dir)
    bidslist = layout.get(return_type='file',subject=participant_label,session=participant_session,datatype=datatype,suffix=suffix,extension=extension)

    return bidslist


from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging
import json
import sys

IFLOGGER=nlogging.getLogger('nipype.interface')

IS_PRESENT="^^^"
IGNORE="###"

def qsiprep_proc(labels_dict,bids_dir=""):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)

    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["SINGULARITYENV_TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME

    command_base, container = getContainer(labels_dict,nodename="qsiprep", SPECIFIC="QSIPREP_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the qsiprep version:")
    command = f"{command_base} --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    eddy_config = getParams(labels_dict,'EDDY_CONFIG')
    eddy_json=""
    if eddy_config:
        if os.path.exists(eddy_config):
            with open(eddy_config, 'r') as infile:
                eddy_json = json.load(infile)
        if eddy_json:
            IFLOGGER.info(f"eddy params provided in file {eddy_config} and contents are:")
            IFLOGGER.info(f"{eddy_json}")


    # set up dwi to process just the specific dwi session
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    participant_session = getParams(labels_dict,'PARTICIPANT_SESSION')

    # This is for 1 specific scenario - we will terminate early if field map is missing
    layout = BIDSLayout(bids_dir)
    epi = layout.get(subject=participant_label, session=participant_session, suffix="epi",extension=".nii.gz") 
    if not epi:
        IFLOGGER.warn(f"Fieldmap not found for subject {participant_label} and session {participant_session}. terminating script early.")
        sys.exit(1)


    bids_filter_dict={}
    bids_filter_dict["dwi"] = {}
    bids_filter_dict["dwi"]["session"] =  participant_session
    bids_filter_file = os.path.join(cwd,f"{participant_label}_{participant_session}_bids_filter_file.json")
    export_labels(bids_filter_dict,bids_filter_file)
    IFLOGGER.info(f"Specifying session filter: exporting {bids_filter_dict} to {bids_filter_file}")

    qsiprep_dict={}
    qsiprep_dict = updateParams(qsiprep_dict,"--participant_label","<PARTICIPANT_LABEL>")
    qsiprep_dict = updateParams(qsiprep_dict,"--separate-all-dwis",IS_PRESENT)
    qsiprep_dict = updateParams(qsiprep_dict,"--bids-filter-file",bids_filter_file)
    qsiprep_dict = updateParams(qsiprep_dict,"--skip-bids-validation",IS_PRESENT)
    qsiprep_dict = updateParams(qsiprep_dict,"--hmc-model" ,"eddy")
    qsiprep_dict = updateParams(qsiprep_dict,"--unringing-method" ,"rpg")
    qsiprep_dict = updateParams(qsiprep_dict,"--b1-biascorrect-stage","none")
    qsiprep_dict = updateParams(qsiprep_dict,"--eddy-config" ,"<EDDY_CONFIG>")
    qsiprep_dict = updateParams(qsiprep_dict,"--mem_mb","<BIDSAPP_MEMORY>")
    qsiprep_dict = updateParams(qsiprep_dict,"--nthreads","<BIDSAPP_THREADS>")
    qsiprep_dict = updateParams(qsiprep_dict,"--fs-license-file","<FSLICENSE>")
    qsiprep_dict = updateParams(qsiprep_dict,"--write-graph",IS_PRESENT)
    qsiprep_dict = updateParams(qsiprep_dict,"--output-resolution","<OUTPUT_RES>")
    qsiprep_dict = updateParams(qsiprep_dict,"-w","<CWD>/qsiprep_work")


    # Additional params
    QSIPREP_OVERRIDE_PARAMS = getParams(labels_dict,"QSIPREP_OVERRIDE_PARAMS")
    if QSIPREP_OVERRIDE_PARAMS and isinstance(QSIPREP_OVERRIDE_PARAMS,dict):
        add_labels(QSIPREP_OVERRIDE_PARAMS,qsiprep_dict)        

    params = ""
    for qsiprep_tag, qsiprep_value in qsiprep_dict.items():
        if "--" in qsiprep_tag and "---" not in qsiprep_tag:
            if qsiprep_value == IS_PRESENT:
                params=params + " " + qsiprep_tag
            elif qsiprep_value == IGNORE:
                IFLOGGER.info(f"Parameter {qsiprep_tag} is being skipped. This has been explicitly required in configuration.")
            else:
                # we dont need = sign for qsiprep just for basi;
                params = params + " " + qsiprep_tag + " " + qsiprep_value

        elif "-" in qsiprep_tag and "--" not in qsiprep_tag:
            params = params + " " + qsiprep_tag + " " + qsiprep_value

        else:
            print(f"qsiprep tag {qsiprep_tag} not valid.") 


    command=f"{command_base}"\
            " "+bids_dir +\
            " <CWD>"\
            " participant"\
            " " + params

    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    dwi_preprocess = getGlob(os.path.join(cwd,'qsiprep','sub-{}'.format(participant_label),'ses-*','dwi','*preproc_dwi.nii.gz'))
    mat_t12mni = getGlob(os.path.join(cwd,'qsiprep','sub-{}'.format(participant_label),'anat','*from-T1w*mode-image_xfm.h5'))
    mat_mni2t1 = getGlob(os.path.join(cwd,'qsiprep','sub-{}'.format(participant_label),'anat','*from-MNI*mode-image_xfm.h5'))
    t1ref =  getGlob(os.path.join(cwd,'qsiprep','sub-{}'.format(participant_label),'anat','*{}_desc-preproc_T1w.nii.gz'.format(participant_label)))
    mniref =  getGlob(os.path.join(cwd,'qsiprep','sub-{}'.format(participant_label),'anat','*{}*MNI*desc-preproc_T1w.nii.gz'.format(participant_label)))
    output_dir = cwd

    
    out_files=[]
    out_files.insert(0,dwi_preprocess)
    out_files.insert(1,mat_t12mni)
    out_files.insert(2,mat_mni2t1)
    out_files.insert(3,mniref)
    out_files.insert(4,t1ref)


    return {
        "dwi_preprocess":dwi_preprocess,
        "mat_t12mni":mat_t12mni,
        "mat_mni2t1":mat_mni2t1,
        "mniref":mniref,
        "t1ref":t1ref,
        "output_dir":output_dir,
        "out_files":out_files
    }



class qsiprepInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class qsiprepOutputSpec(TraitedSpec):
    dwi_preprocess = File(desc='Preprocessed DWI')
    mat_t12mni = File(desc='T1 to MNI transform')
    mat_mni2t1 = File(desc='MNI to T1 transform')
    mniref = File(desc='MNI reference')
    t1ref = File(desc='T1 reference')
    output_dir = traits.String(desc="QSIPREP output directory")
    out_files = traits.List(desc='list of files')
    
class qsiprep_pan(BaseInterface):
    input_spec = qsiprepInputSpec
    output_spec = qsiprepOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = qsiprep_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='qsiprep_node',bids_dir="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(qsiprep_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





from nipype import Workflow

import panpipelines.nodes.qsiprep as qsiprep
from panpipelines.utils.util_functions import *


def create(name, wf_base_dir,labels_dict,createGraph=True,execution={},LOGGER=None):
    # Create workflow
    pan_workflow = Workflow(name=name, base_dir=wf_base_dir)

    if LOGGER:
        LOGGER.info(f"Created Workflow {name} with base directory {wf_base_dir}")

    if len(execution.keys()) > 0:
        pan_workflow.config = process_dict(pan_workflow.config,execution)

    # Specify node inputs
    qsiprep_node = qsiprep.create(labels_dict,LOGGER=LOGGER)

    pan_workflow.add_nodes([qsiprep_node])

    if createGraph:
         pan_workflow.write_graph(graph2use='flat')


    return pan_workflow



from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import numpy as np 
import nibabel as nib
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def parse_textdata_proc(labels_dict, textdata, textdata_type):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    
    output_dir=cwd
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')
    roi_output_dir = os.path.join(cwd,'{}_roi_output_dir'.format(participant_label))

    NOSESSION=False
    if not session_label:
        session_label = "NotProvided"
        NOSESSION=True
    else:
        roi_output_dir = os.path.join(cwd,'{}_{}_roi_output_dir'.format(participant_label,session_label))

    if not os.path.isdir(roi_output_dir):
        os.makedirs(roi_output_dir)

    out_files=[]
    roi_csv = None
    df=None
    basefile_name = os.path.basename(textdata)
    IFLOGGER.info(f"Data file {textdata} provided. Using {basefile_name} for csv output base.")


    pipeline_name = getParams(labels_dict,'PIPELINE')
    if not pipeline_name:
        pipeline_name="UnDetermined"

    # Need to think about this additional prefix - only included it because of the use case of comparing
    # different freesurfer pipelines. There should be a more elegant way to handle this.
    # Actually yes - just get the pipeline name or desc!
    subject_project = getParams(labels_dict,'PARTICIPANT_XNAT_PROJECT')
    if subject_project:
        creating_pipeline = textdata.split("/" + subject_project)[0].split("/")[-1]
    else:
        creating_pipeline=""
        
    addpref=""
    if creating_pipeline:
        addpref = creating_pipeline + "-"

    if "aseg" in basefile_name or textdata_type=="aseg":
        prefix= addpref + "aseg"
        df = get_freesurfer_genstats(textdata,columns=["Volume_mm3"], prefix=prefix,participant_label=participant_label,session_label=session_label)
    elif "lh.aparc.a2009s" in basefile_name or textdata_type=="lh.aparc.a2009s":
        prefix= addpref + "lh-Destrieux"
        df = get_freesurfer_genstats(textdata,columns=["SurfArea","GrayVol","ThickAvg"], prefix=prefix,participant_label=participant_label,session_label=session_label)
    elif "rh.aparc.a2009s" in basefile_name or textdata_type=="rh.aparc.a2009s":
        prefix= addpref + "rh-Destrieux"
        df = get_freesurfer_genstats(textdata,columns=["SurfArea","GrayVol","ThickAvg"], prefix=prefix,participant_label=participant_label,session_label=session_label)
    elif "lh.aparc" in basefile_name or textdata_type=="lh.aparc":
        prefix= addpref + "lh-DK"
        df = get_freesurfer_genstats(textdata,columns=["SurfArea","GrayVol","ThickAvg"], prefix=prefix,participant_label=participant_label,session_label=session_label)
    elif "rh.aparc" in basefile_name or textdata_type=="rh.aparc":
        prefix= addpref + "rh-DK"
        df = get_freesurfer_genstats(textdata,columns=["SurfArea","GrayVol","ThickAvg"], prefix=prefix,participant_label=participant_label,session_label=session_label)     
    elif "hipposubfields.lh" in basefile_name or textdata_type=="hipposubfields.lh":
        prefix= addpref + "lh-hipposf" + basefile_name.split("hipposubfields.lh.")[1].split(".stats")[0].replace(".","-")
        df = get_freesurfer_hippostats(textdata,prefix=prefix, participant_label=participant_label,session_label=session_label)
    elif "hipposubfields.rh" in basefile_name or textdata_type=="hipposubfields.rh":
        prefix= addpref + "rh-hipposf" + basefile_name.split("hipposubfields.rh.")[1].split(".stats")[0].replace(".","-")
        df = get_freesurfer_hippostats(textdata,prefix=prefix, participant_label=participant_label,session_label=session_label)
    elif "hippoSfVolumes" in basefile_name or textdata_type=="hippoSfVolumes":
        prefix =  addpref + basefile_name.split(".hippoSfVolumes")[0] + "-hippo" + basefile_name.split(".hippoSfVolumes")[1].split(".txt")[0].replace(".","-")
        df = get_freesurfer_subregionstats(textdata,prefix=prefix, participant_label=participant_label,session_label=session_label)
    elif "amygNucVolumes" in basefile_name or textdata_type=="amygNucVolumes":
        prefix =  addpref + basefile_name.split(".amygNucVolumes")[0] + "-amyg" + basefile_name.split(".amygNucVolumes")[1].split(".txt")[0].replace(".","-")
        df = get_freesurfer_subregionstats(textdata,prefix=prefix, participant_label=participant_label,session_label=session_label)
    elif "ThalamicNuclei" in basefile_name or textdata_type=="ThalamicNuclei":
        prefix =  addpref + "thalamic" + basefile_name.split("ThalamicNuclei")[1].split(".volumes.txt")[0].replace(".","-")
        df = get_freesurfer_subregionstats(textdata,prefix=prefix, participant_label=participant_label,session_label=session_label)
    elif "brainstemSsLabelsbeta" in basefile_name or textdata_type=="brainstembeta":
        prefix =  addpref + "brainstembeta"
        df = get_freesurfer_subregionstats(textdata,prefix=prefix, participant_label=participant_label,session_label=session_label)
    else:
        IFLOGGER.info("basename {textdata} not implemented.")


    if df is not None:
        if NOSESSION:
            roi_csv = os.path.join(roi_output_dir,'{}_{}_{}.csv'.format(participant_label,pipeline_name,basefile_name))
        else:
            roi_csv = os.path.join(roi_output_dir,'{}_{}_{}_{}.csv'.format(participant_label,session_label,pipeline_name,basefile_name))            
        df.to_csv(roi_csv,sep=",",header=True, index=False)
        out_files.insert(0,roi_csv)

        metadata = {}
        roi_csv_json = os.path.splitext(roi_csv)[0] + ".json"
        metadata = updateParams(metadata,"Title","parse_textdata.py")
        metadata = updateParams(metadata,"Description","Parse textdata from existing tables into csv using predefined approaches. Freesurfer stat files are supported.")
        metadata = updateParams(metadata,"MetadataFile",f"{roi_csv_json}")
        metadata = updateParams(metadata,"FileCreated",f"{roi_csv}")
        metadata = updateParams(metadata,"DateCreated",datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S:%f"))
        metadata = updateParams(metadata,"Pipeline",f"{pipeline_name}")
        metadata = updateParams(metadata,"InputFile",f"{textdata}")
        metadata = updateParams(metadata,"InputFilePipeline",f"{creating_pipeline}")
        export_labels(metadata,roi_csv_json)

    return {
        "roi_csv":roi_csv,
        "roi_output_dir":roi_output_dir,
        "output_dir":output_dir,
        "out_files":out_files
    }



class parse_textdataInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    textdata = File(desc='text data file to parse')
    textdata_type = traits.String(desc='type of file to help discern parsing method')

class parse_textdataOutputSpec(TraitedSpec):
    roi_csv = File(desc='CSV file of results')
    roi_output_dir = traits.String(desc='roi output dir')
    output_dir = traits.String(desc='output dir')
    out_files = traits.List(desc='list of files')
    
class parse_textdata_pan(BaseInterface):
    input_spec = parse_textdataInputSpec
    output_spec = parse_textdataOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = parse_textdata_proc(
            self.inputs.labels_dict,
            self.inputs.textdata,
            self.inputs.textdata_type
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="parse_textdata_node",textdata="",textdata_type="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(parse_textdata_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
            
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if not textdata is None and not textdata == "":
        pan_node.inputs.textdata = textdata
 
    if not textdata_type is None and not textdata_type == "":
        pan_node.inputs.textdata_type = textdata_type                

    return pan_node





from panpipelines.utils.util_functions import *
from panpipelines.scripts.panscript import *
import os
import glob

# TEST
#from panpipelines.scripts import *
#SCRIPT="fmriprep_panscript"
#panscript=eval("{}.{}".format(SCRIPT,SCRIPT))
#labels_dict={"COMM": "ls"}
#pancomm = panscript(labels_dict)
#pancomm.run()


class fmriprep_panscript(panscript):

    def __init__(self,labels_dict,name='fmriprep_panscript',params="",command="",execution={}):
        super().__init__(self,labels_dict,name=name,params=params,command=command)

        self.params = "--participant_label <PARTICIPANT_LABEL>" \
            " --output-spaces MNI152NLin6Asym:res-2 MNI152NLin2009cAsym:res-2 fsLR fsaverage anat func"\
            " --skip-bids-validation"\
            " --mem_mb <BIDSAPP_MEMORY>" \
            " --nthreads <BIDSAPP_THREADS>"\
            " --fs-license-file <FSLICENSE>"\
            " --omp-nthreads <BIDSAPP_THREADS>"\
            " -w <OUTPUT_DIR>/fmriwork"

        self.command = "singularity run --cleanenv --nv --no-home <FMRIPREP_CONTAINER>"\
                " <BIDS_DIR>"\
                " <OUTPUT_DIR>/fmrioutput"\
                " participant"

    def pre_run(self):
        print("pre run - setting template flow directory")
        TEMPLATEFLOW_HOME=getParams(self.labels_dict,"TEMPLATEFLOW_HOME")
        os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
        os.environ["SINGULARITYENV_TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME

        print("Creating output directory")
        output_dir = getParams(self.labels_dict,"OUTPUT_DIR")
        if os.path.exists(output_dir):
            os.makedirs(output_dir)
    
    def post_run(self):
        print("post run")

    def get_results(self):
        self.results = {}
        return self.results








from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
from panpipelines.utils.transformer import *
import os
import glob
import nibabel as nb
import pathlib
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def antstransform_proc(labels_dict,input_file,trans_mat,ref_file):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    output_dir = cwd

    command_base, container = getContainer(labels_dict,nodename="antstransform",SPECIFIC="ANTS_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the ants version:")
    command = f"{command_base} antsRegistration --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    if container:
        IFLOGGER.info("\nChecking the container version:")
        command = f"{command_base} --version"
        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

    if Path(input_file).suffix == ".mgz":
        mgzdir = os.path.join(cwd,'mgz_nii')
        if not os.path.isdir(mgzdir):
            os.makedirs(mgzdir)

        fs_command_base, fscontainer = getContainer(labels_dict,nodename="convMGZ2NII",SPECIFIC="FREESURFER_CONTAINER",LOGGER=IFLOGGER)
        input_file_nii = newfile(mgzdir,input_file,extension=".nii.gz")
        convMGZ2NII(input_file, input_file_nii, fs_command_base)
        input_file = input_file_nii

    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    work_dir = os.path.join(cwd,'{}_workdir'.format(participant_label))
    if not os.path.isdir(work_dir):
        os.makedirs(work_dir)

    if isinstance(trans_mat[0],list):
        trans_parts = substitute_labels(trans_mat[0][-1],labels_dict).split(":")
    else:
        trans_parts = substitute_labels(trans_mat[-1],labels_dict).split(":")
    if len(trans_parts)>3:
        trans_mat_last = trans_parts[3]
    else:
        trans_mat_last = trans_parts[0]

    trans_mat_basename = os.path.basename(trans_mat_last)
    find_new_space = trans_mat_basename.split("_to-")
    if len(find_new_space) > 1:
        trans_space="space-"+trans_mat_basename.split("_to-")[1].split("_")[0] 
    else:
        trans_space=trans_mat_basename.split(".")[0]


    input_file_basename = os.path.basename(input_file)
    find_old_space=input_file_basename.split("_space-")
    if len(find_old_space) > 1:
        old_space="space-" + input_file_basename.split("_space-")[1].split("_")[0].split('.')[0] 
    else:
        old_space=None

    
    if old_space:
        out_file = os.path.join(output_dir,input_file_basename.replace(old_space, trans_space))
    else:
        out_file = newfile(output_dir,input_file_basename,intwix=trans_space)

    #ensure extension set to nifti
    out_file = newfile(assocfile=out_file,extension=".nii.gz")

    costfunction = getParams(labels_dict,'COST_FUNCTION')
    if costfunction is None:
        costfunction="LanczosWindowedSinc"

    output_type = getParams(labels_dict,'OUTPUT_TYPE')


    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")


    transform_list=[]
    reverse_list=[]
    trans_ori = "RAS"

    for trans in trans_mat:
        # check if list passed within list then force to be single string item
        if isinstance(trans,list):
            trans = trans[0]
        trans_parts = trans.split(":")
        transform = getGlob(substitute_labels(trans_parts[0],labels_dict))
        trans_type =""
        trans_source = ""
        trans_reference = ""
        trans_reverse =  ""

        if len(trans_parts) == 6:
            trans_type = trans_parts[1]
            trans_source = substitute_labels(trans_parts[2],labels_dict)
            trans_reference = substitute_labels(trans_parts[3],labels_dict)
            trans_reverse = trans_parts[4]
            trans_ori = trans_parts[5]
        elif len(trans_parts) == 5:
            trans_type = trans_parts[1]
            trans_source = substitute_labels(trans_parts[2],labels_dict)
            trans_reference = substitute_labels(trans_parts[3],labels_dict)
            trans_reverse = trans_parts[4]
        elif len(trans_parts) == 4:
            trans_type = trans_parts[1]
            trans_source = substitute_labels(trans_parts[2],labels_dict)
            trans_reference = substitute_labels(trans_parts[3],labels_dict)
        elif len(trans_parts) == 3:
            trans_type = trans_parts[1]
            trans_source = substitute_labels(trans_parts[2],labels_dict)
        elif len(trans_parts) == 2:
            trans_type = trans_parts[1]

        if trans_reverse:
            if trans_reverse == "True":
                reverse_list.append(True)
            else:
                reverse_list.append(False)

        else:
            reverse_list.append(False)

        TRANSLIT="from-MNI152NLin6Asym_to-MNI152NLin2009cAsym"
        if TRANSLIT in str(transform):
            # if ref file not defined then assume that user implies this as reference
            if not ref_file:
                ref_file = TRANSLIT.split("_to-")[1]

            transform = get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin2009cAsym",suffix="xfm",extension=[".h5"])

        TRANSLIT="from-MNI152NLin2009cAsym_to-MNI152NLin6Asym"
        if TRANSLIT in str(transform):
            # if ref file not defined then assume that user implies this as reference
            if not ref_file:
                ref_file = TRANSLIT.split("_to-")[1]

            transform = get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin6Asym",suffix="xfm",extension=[".h5"])

        if transform == "tkregister2_fslout":
            new_freesurfer_transform = newfile(work_dir,transform,suffix="fsl-transform")
            fs_command_base, fscontainer = getContainer(labels_dict,nodename="convMGZ2NII",SPECIFIC="FREESURFER_CONTAINER",LOGGER=IFLOGGER)
            tkregister2_fslout(trans_source,trans_reference, fs_command_base,new_freesurfer_transform)
            transform = new_freesurfer_transform

        if trans_type == "FSL":
            
            if pathlib.Path(transform).suffix == ".gz":
                # if we need the inverse of non-linear transform FSL transform then do that first before converting to ANTS and then reset reverse:
                if reverse_list[-1]==True:
                    new_transform = newfile(work_dir,transform,suffix="desc-inverse",extension=".nii.gz")
                    fsl_command_base, fslcontainer = getContainer(labels_dict,nodename="invertWarpfield_FNIRT",SPECIFIC="FSL_CONTAINER",LOGGER=IFLOGGER)
                    invertWarpfield_FNIRT(trans_source, transform, new_transform ,fsl_command_base)
                    transform = new_transform
                    reverse_list[-1]=False
                    trans_source = trans_reference

                new_ants_transform=newfile(work_dir,transform,suffix="ants-transform", extension=".nii.gz")
                wb_command_base, wbcontainer = getContainer(labels_dict,nodename="convertwarp_toANTS",SPECIFIC="WB_CONTAINER",LOGGER=IFLOGGER)
                convertwarp_toANTS(transform,trans_source, new_ants_transform, wb_command_base )
            else:
                new_ants_transform=newfile(work_dir,transform,suffix="ants-transform", extension=".mat")
                convert_affine_fsl_to_ants(transform, trans_source, trans_reference, new_ants_transform)
            transform = new_ants_transform
        
        transform_list.append(transform)

    # Process Reference_file
    ref_parts = ref_file.split(":")
    ref_file = getGlob(ref_parts[0])
    new_ref_file = newfile(work_dir,ref_file,suffix="desc-resample")

    REFLIT="MNI152NLin2009cAsym_res-"
    if REFLIT in str(ref_file):
        resolution=int(ref_file.split(REFLIT)[1])
        ref_file=get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin2009cAsym",resolution=resolution,suffix="T1w",extension=[".nii.gz"])

    REFLIT="MNI152NLin6Asym_res-"
    if REFLIT in str(ref_file):
        resolution=int(ref_file.split(REFLIT)[1])
        ref_file=get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin6Asym",resolution=resolution,suffix="T1w",extension=[".nii.gz"])

    if len(ref_parts) == 5:
        newdims_parts = ref_parts[1].split("|")
        newdims = newdims_parts[0]
        if len(newdims_parts) > 1:
            dim_type = newdims_parts[1]
        else:
            dim_type = ""
        ref_ori = ref_parts[2]
        interpolation_type = ref_parts[3]
        output_type = ref_parts[4]
        IFLOGGER.info("Calling function resampleimage_ants_ori with parameters:")
        IFLOGGER.info(f"input_file: {ref_file}")
        IFLOGGER.info(f"out_file: {new_ref_file}")
        IFLOGGER.info(f"newdims: {newdims}")
        IFLOGGER.info(f"dim_type: {dim_type}")
        IFLOGGER.info(f"ref_ori: {ref_ori}")
        IFLOGGER.info(f"interpolation_type: {interpolation_type}")
        IFLOGGER.info(f"output_type: {output_type:}")
        new_ref_file = resampleimage_ants_ori(ref_file,new_ref_file,newdims,command_base,dim_type=dim_type, target_ori=ref_ori,interpolation_type=interpolation_type,output_type=output_type)

    elif len(ref_parts) == 4:
        newdims_parts = ref_parts[1].split("|")
        newdims = newdims_parts[0]
        if len(newdims_parts) > 1:
            dim_type = newdims_parts[1]
        else:
            dim_type = ""
        ref_ori = ref_parts[2]
        interpolation_type = ref_parts[3]
        IFLOGGER.info("Calling function resampleimage_ants_ori with parameters:")
        IFLOGGER.info(f"input_file: {ref_file}")
        IFLOGGER.info(f"out_file: {new_ref_file}")
        IFLOGGER.info(f"newdims: {newdims}")
        IFLOGGER.info(f"dim_type: {dim_type}")
        IFLOGGER.info(f"ref_ori: {ref_ori}")
        IFLOGGER.info(f"interpolation_type: {interpolation_type}")
        new_ref_file = resampleimage_ants_ori(ref_file,new_ref_file,newdims,command_base,dim_type=dim_type,target_ori=ref_ori,interpolation_type=interpolation_type)
    
    elif len(ref_parts) == 3:
        newdims_parts = ref_parts[1].split("|")
        newdims = newdims_parts[0]
        if len(newdims_parts) > 1:
            dim_type = newdims_parts[1]
        else:
            dim_type = ""
        ref_ori = ref_parts[2]
        IFLOGGER.info("Calling function resampleimage_ants_ori with parameters:")
        IFLOGGER.info(f"input_file: {ref_file}")
        IFLOGGER.info(f"out_file: {new_ref_file}")
        IFLOGGER.info(f"newdims: {newdims}")
        IFLOGGER.info(f"dim_type: {dim_type}")
        IFLOGGER.info(f"ref_ori: {ref_ori}")
        new_ref_file = resampleimage_ants_ori(ref_file,new_ref_file,newdims,command_base,dim_type=dim_type,target_ori=ref_ori)

    elif len(ref_parts) == 2:
        newdims_parts = ref_parts[1].split("|")
        newdims = newdims_parts[0]
        if len(newdims_parts) > 1:
            dim_type = newdims_parts[1]
        else:
            dim_type = ""
        IFLOGGER.info("Calling function resampleimage_ants_ori with parameters:")
        IFLOGGER.info(f"input_file: {ref_file}")
        IFLOGGER.info(f"out_file: {new_ref_file}")
        IFLOGGER.info(f"newdims: {newdims}")
        IFLOGGER.info(f"dim_type: {dim_type}")
        new_ref_file = resampleimage_ants_ori(ref_file,new_ref_file,newdims,command_base,dim_type=dim_type,)

    if os.path.exists(new_ref_file):
        ref_file = new_ref_file

    IFLOGGER.info("Calling function apply_transform_ants_ori with parameters:")
    IFLOGGER.info(f"input_file: {input_file}")
    IFLOGGER.info(f"ref_file: {ref_file}")
    IFLOGGER.info(f"out_file: {out_file}")
    IFLOGGER.info(f"transform_list: {transform_list}")
    IFLOGGER.info(f"target_ori: {trans_ori}")
    IFLOGGER.info(f"costfunction: {costfunction}")
    IFLOGGER.info(f"output_type: {output_type}")
    IFLOGGER.info(f"reverse_list: {reverse_list}")

    apply_transform_ants_ori(input_file,
                            ref_file,
                            out_file,
                            transform_list,
                            command_base,
                            target_ori=trans_ori,                                              
                            costfunction=costfunction,
                            output_type=output_type,
                            reverse=reverse_list)

    out_files=[]
    out_files.insert(0,out_file)


    return {
        "out_file":out_file,
        "output_dir":output_dir,
        "out_files":out_files
    }



class antstransformInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    input_file = File(mandatory=True,desc="Image file to transform")
    trans_mat = traits.List(desc='list of transforms')
    ref_file = File(mandatory=False,desc="Image file to transform")

class antstransformOutputSpec(TraitedSpec):
    out_file = File(desc='transformed file')
    output_dir = traits.String(desc="Transform output directory")
    out_files = traits.List(desc='list of files')
    
class antstransform_pan(BaseInterface):
    input_spec = antstransformInputSpec
    output_spec = antstransformOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = antstransform_proc(
            self.inputs.labels_dict,
            self.inputs.input_file,
            self.inputs.trans_mat,
            self.inputs.ref_file
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="antstransform_node",input_file="",trans_mat="",ref_file="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(antstransform_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")

    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    pan_node.inputs.input_file =  input_file       
    pan_node.inputs.trans_mat =  trans_mat
    pan_node.inputs.ref_file =  ref_file

    return pan_node





from nipype import Workflow, MapNode, Node

import panpipelines.nodes.antstransform as antstransform
import panpipelines.nodes.collate_csv_group as collate_csv_group
from panpipelines.utils.util_functions import *
from panpipelines.utils.transformer import *
import glob

def create(name, wf_base_dir,labels_dict,createGraph=True,execution={}, LOGGER=None):
    # Create workflow
    pan_workflow = Workflow(name=name, base_dir=wf_base_dir)

    if LOGGER:
        LOGGER.info(f"Created Workflow {name} with base directory {wf_base_dir}")

    if len(execution.keys()) > 0:
        pan_workflow.config = process_dict(pan_workflow.config,execution)

    # we will not evaluate these variables now as we will need them to be evaluated later in the node
    EXCEPTIONS=["PARTICIPANT_LABEL","PARTICIPANT_XNAT_RPOJECT"]

    measures_list1=[]
    measures_template = getParams(labels_dict,"MEASURES_TEMPLATE1")
    if isinstance(measures_template,list):
        for meas_template in measures_template:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict, EXCEPTIONS)
            measures_list1.extend([evaluated_meas_template])
    elif measures_template:
        measures_list1.extend([substitute_labels(meas_template,labels_dict, EXCEPTIONS)])

    measures_list1.sort()

    measures_list2=[]
    measures_template = getParams(labels_dict,"MEASURES_TEMPLATE2")
    if isinstance(measures_template,list):
        for meas_template in measures_template:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict, EXCEPTIONS)
            measures_list2.extend([evaluated_meas_template])
    elif measures_list2:
        measures_list2.extend([substitute_labels(meas_template,labels_dict, EXCEPTIONS)])

    measures_list2.sort()

    collate_csv_groupnode = collate_csv_group.create(labels_dict, csv_list1=measures_list1, csv_list2=measures_list2, LOGGER=LOGGER)
    pan_workflow.add_nodes([collate_csv_groupnode])


    if createGraph:
         pan_workflow.write_graph(graph2use='flat')


    return pan_workflow



from panpipelines.utils.util_functions import *
from subprocess import run
import os
import glob


class panpipeline:

    def __init__(self, labels_dict,pipeline_dir, participant_label, name='template_panpipeline', createGraph=True, LOGGER=None, execution={},analysis_level="participant", participant_project=None, participant_session=None):
        self.labels_dict = labels_dict
        self.pipeline_dir = pipeline_dir
        self.participant_label = participant_label
        self.name = name
        self.createGraph=createGraph
        self.execution = execution
        self.LOGGER = LOGGER
        self.analysis_level = analysis_level
        self.participant_project = participant_project
        self.participant_session = participant_session
        self.results={}


    def pre_run(self):
        log_message=f"Running {self.name} pipeline at {self.analysis_level} level for "
        if self.analysis_level == "participant":
            log_message = log_message + f"participant {self.participant_label} in "
            if self.participant_session:
                log_message = log_message + f"session {self.participant_session} in "
                if self.participant_project:
                    log_message = log_message + f"project {self.participant_project}\n"
        else:
            participant_info_pairing = [ it for it in zip(self.participant_label,self.participant_project, self.participant_session)]
            for part_info in participant_info_pairing:
                log_message = log_message + f"participant {part_info[0]} in "
                if part_info[1]:
                    log_message = log_message + f"session {part_info[1]} in "
                    if part_info[2]:
                        log_message = log_message + f"project {part_info[2]}\n" 
        if self.LOGGER:
            self.LOGGER.info(log_message)     

    def proc(self):
        pass

    def run(self):

        self.pre_run()

        self.proc()

        self.post_run()

        results = self.get_results()
        if self.LOGGER:
            if results:
                self.LOGGER.info(results)
        return results

    def post_run(self):
        log_message=f"Completed {self.name} pipeline at {self.analysis_level} level for "
        if self.analysis_level == "participant":
            log_message = log_message + f"participant {self.participant_label} in "
            if self.participant_session:
                log_message = log_message + f"session {self.participant_session} in "
                if self.participant_project:
                    log_message = log_message + f"project {self.participant_project} "
        else:
            participant_info_pairing = [ it for it in zip(self.participant_label,self.participant_project, self.participant_session)]
            for part_info in participant_info_pairing:
                log_message = log_message + f"participant {part_info[0]} in "
                if part_info[1]:
                    log_message = log_message + f"session {part_info[1]} in "
                    if part_info[2]:
                        log_message = log_message + f"project {part_info[2]} \n" 
        if self.LOGGER:
            self.LOGGER.info(log_message)          


    def get_results(self):
        return self.results





